# Copyright (c) 2013 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import collections
from functools import reduce


class Graph(object):
    """A standard graph data structure.

    With routines applicable to analysis of policy.
    """
    class dfs_data(object):
        """Data for each node in graph during depth-first-search."""
        def __init__(self, begin=None, end=None):
            self.begin = begin
            self.end = end

        def __str__(self):
            return "<begin: %s, end: %s>" % (self.begin, self.end)

    class edge_data(object):
        """Data for each edge in graph."""
        def __init__(self, node=None, label=None):
            self.node = node
            self.label = label

        def __str__(self):
            return "<Label:%s, Node:%s>" % (self.label, self.node)

        def __eq__(self, other):
            return self.node == other.node and self.label == other.label

        def __hash__(self):
            return hash(str(self))

    def __init__(self, graph=None):
        self.edges = {}   # dict from node to list of nodes
        self.nodes = {}   # dict from node to info about node
        self._cycles = None

    def __or__(self, other):
        # do this the simple way so that subclasses get this code for free
        g = self.__class__()
        for node in self.nodes:
            g.add_node(node)
        for node in other.nodes:
            g.add_node(node)

        for name in self.edges:
            for edge in self.edges[name]:
                g.add_edge(name, edge.node, label=edge.label)
        for name in other.edges:
            for edge in other.edges[name]:
                g.add_edge(name, edge.node, label=edge.label)
        return g

    def __ior__(self, other):
        if len(other) == 0:
            # no changes if other is empty
            return self
        self._cycles = None
        for name in other.nodes:
            self.add_node(name)
        for name in other.edges:
            for edge in other.edges[name]:
                self.add_edge(name, edge.node, label=edge.label)
        return self

    def __len__(self):
        return (len(self.nodes) +
                reduce(lambda x, y: x+y,
                       (len(x) for x in self.edges.values()),
                       0))

    def add_node(self, val):
        """Add node VAL to graph."""
        if val not in self.nodes:  # preserve old node info
            self.nodes[val] = None
            return True
        return False

    def delete_node(self, val):
        """Delete node VAL from graph and all edges."""
        try:
            del self.nodes[val]
            del self.edges[val]
        except KeyError:
            assert val not in self.edges

    def add_edge(self, val1, val2, label=None):
        """Add edge from VAL1 to VAL2 with label LABEL to graph.

        Also adds the nodes.
        """
        self._cycles = None  # so that has_cycles knows it needs to rerun
        self.add_node(val1)
        self.add_node(val2)
        val = self.edge_data(node=val2, label=label)
        try:
            self.edges[val1].add(val)
        except KeyError:
            self.edges[val1] = set([val])

    def delete_edge(self, val1, val2, label=None):
        """Delete edge from VAL1 to VAL2 with label LABEL.

        LABEL must match (even if None).  Does not delete nodes.
        """
        try:
            edge = self.edge_data(node=val2, label=label)
            self.edges[val1].remove(edge)
        except KeyError:
            # KeyError either because val1 or edge
            return
        self._cycles = None

    def node_in(self, val):
        return val in self.nodes

    def edge_in(self, val1, val2, label=None):
        return (val1 in self.edges and
                self.edge_data(val2, label) in self.edges[val1])

    def reset_nodes(self):
        for node in self.nodes:
            self.nodes[node] = None

    def depth_first_search(self, roots=None):
        """Run depth first search on the graph.

        Also modify self.nodes, self.counter, and self.cycle.
        Use all nodes if @roots param is None or unspecified
        """
        self.reset()
        if roots is None:
            roots = self.nodes
        for node in roots:
            if node in self.nodes and self.nodes[node].begin is None:
                self.dfs(node)

    def _enumerate_cycles(self):
        self.reset()
        for node in self.nodes.keys():
            self._reset_dfs_data()
            self.dfs(node, target=node)
            for path in self.__target_paths:
                self._cycles.add(Cycle(path))

    def reset(self, roots=None):
        """Return nodes to pristine state."""
        self._reset_dfs_data()
        roots = roots or self.nodes
        self._cycles = set()

    def _reset_dfs_data(self):
        for node in self.nodes.keys():
            self.nodes[node] = self.dfs_data()
        self.counter = 0
        self.__target_paths = []

    def dfs(self, node, target=None, dfs_stack=None):
        """DFS implementation.

        Assumes data structures have been properly prepared.
        Creates start/begin times on nodes.
        Adds paths from node to target to self.__target_paths
        """
        if dfs_stack is None:
            dfs_stack = []
        dfs_stack.append(node)
        if (target is not None and node == target and
                len(dfs_stack) > 1):  # non-trival path to target found
            self.__target_paths.append(list(dfs_stack))  # record
        if self.nodes[node].begin is None:
            self.nodes[node].begin = self.next_counter()
            if node in self.edges:
                for edge in self.edges[node]:
                    self.dfs(edge.node, target=target, dfs_stack=dfs_stack)
            self.nodes[node].end = self.next_counter()
        dfs_stack.pop()

    def stratification(self, labels):
        """Return the stratification result.

        Return mapping of node name to integer indicating the
        stratum to which that node is assigned. LABELS is the list
        of edge labels that dictate a change in strata.
        """
        stratum = {}
        for node in self.nodes:
            stratum[node] = 1
        changes = True
        while changes:
            changes = False
            for node in self.edges:
                for edge in self.edges[node]:
                    oldp = stratum[node]
                    if edge.label in labels:
                        stratum[node] = max(stratum[node],
                                            1 + stratum[edge.node])
                    else:
                        stratum[node] = max(stratum[node],
                                            stratum[edge.node])
                    if oldp != stratum[node]:
                        changes = True
                    if stratum[node] > len(self.nodes):
                        return None
        return stratum

    def roots(self):
        """Return list of nodes with no incoming edges."""
        possible_roots = set(self.nodes)
        for node in self.edges:
            for edge in self.edges[node]:
                if edge.node in possible_roots:
                    possible_roots.remove(edge.node)
        return possible_roots

    def has_cycle(self):
        """Checks if there are cycles.

        Run depth_first_search only if it has not already been run.
        """
        if self._cycles is None:
            self._enumerate_cycles()
        return len(self._cycles) > 0

    def cycles(self):
        """Return list of cycles. None indicates unknown. """
        if self._cycles is None:
            self._enumerate_cycles()
        cycles_list = []
        for cycle_graph in self._cycles:
            cycles_list.append(cycle_graph.list_repr())
        return cycles_list

    def dependencies(self, node):
        """Returns collection of node names reachable from NODE.

        If NODE does not exist in graph, returns None.
        """
        if node not in self.nodes:
            return None
        self.reset()
        node_obj = self.nodes[node]

        if node_obj is None or node_obj.begin is None or node_obj.end is None:
            self.depth_first_search([node])
            node_obj = self.nodes[node]
        return set([n for n, dfs_obj in self.nodes.items()
                    if dfs_obj.begin is not None])

    def next_counter(self):
        """Return next counter value and increment the counter."""
        self.counter += 1
        return self.counter - 1

    def __str__(self):
        s = "{"
        for node in self.nodes:
            s += "(" + str(node) + " : ["
            if node in self.edges:
                s += ", ".join([str(x) for x in self.edges[node]])
            s += "],\n"
        s += "}"
        return s

    def _inverted_edge_graph(self):
        """create a shallow copy of self with the edges inverted"""
        newGraph = Graph()
        newGraph.nodes = self.nodes
        for source_node in self.edges:
            for edge in self.edges[source_node]:
                try:
                    newGraph.edges[edge.node].add(Graph.edge_data(source_node))
                except KeyError:
                    newGraph.edges[edge.node] = set(
                        [Graph.edge_data(source_node)])
        return newGraph

    def find_dependent_nodes(self, nodes):
        """Return all nodes dependent on @nodes.

        Node T is dependent on node T.
        Node T is dependent on node R if there is an edge from node S to T,
            and S is dependent on R.

        Note that node T is dependent on node T even if T is not in the graph
        """
        return (self._inverted_edge_graph().find_reachable_nodes(nodes)
                | set(nodes))

    def find_reachable_nodes(self, roots):
        """Return all nodes reachable from @roots."""
        if len(roots) == 0:
            return set()
        self.depth_first_search(roots)
        result = [x for x in self.nodes if self.nodes[x].begin is not None]
        self.reset_nodes()
        return set(result)


class Cycle(frozenset):
    """An immutable set of 2-tuples to represent a directed cycle

    Extends frozenset, adding a list_repr method to represent a cycle as an
    ordered list of nodes.

    The set representation facilicates identity of cycles regardless of order.
    The list representation is much more readable.
    """
    def __new__(cls, cycle):
        edge_list = []
        for i in range(1, len(cycle)):
            edge_list.append((cycle[i - 1], cycle[i]))
        new_obj = super(Cycle, cls).__new__(cls, edge_list)
        new_obj.__list_repr = list(cycle)  # save copy as list_repr
        return new_obj

    def list_repr(self):
        """Return list-of-nodes representation of cycle"""
        return self.__list_repr


class BagGraph(Graph):
    """A graph data structure with bag semantics for nodes and edges.

    Keeps track of the number of times each node/edge has been inserted.
    A node/edge is removed from the graph only once it has been deleted
    the same number of times it was inserted.  Deletions when no node/edge
    already exist are ignored.
    """
    def __init__(self, graph=None):
        super(BagGraph, self).__init__(graph)
        self._node_refcounts = {}  # dict from node to counter
        self._edge_refcounts = {}  # dict from edge to counter

    def add_node(self, val):
        """Add node VAL to graph."""
        super(BagGraph, self).add_node(val)
        if val in self._node_refcounts:
            self._node_refcounts[val] += 1
        else:
            self._node_refcounts[val] = 1

    def delete_node(self, val):
        """Delete node VAL from graph (but leave all edges)."""
        if val not in self._node_refcounts:
            return
        self._node_refcounts[val] -= 1
        if self._node_refcounts[val] == 0:
            super(BagGraph, self).delete_node(val)
            del self._node_refcounts[val]

    def add_edge(self, val1, val2, label=None):
        """Add edge from VAL1 to VAL2 with label LABEL to graph.

        Also adds the nodes VAL1 and VAL2 (important for refcounting).
        """
        super(BagGraph, self).add_edge(val1, val2, label=label)
        edge = (val1, val2, label)
        if edge in self._edge_refcounts:
            self._edge_refcounts[edge] += 1
        else:
            self._edge_refcounts[edge] = 1

    def delete_edge(self, val1, val2, label=None):
        """Delete edge from VAL1 to VAL2 with label LABEL.

        LABEL must match (even if None).  Also deletes nodes
        whenever the edge exists.
        """
        edge = (val1, val2, label)
        if edge not in self._edge_refcounts:
            return
        self.delete_node(val1)
        self.delete_node(val2)
        self._edge_refcounts[edge] -= 1
        if self._edge_refcounts[edge] == 0:
            super(BagGraph, self).delete_edge(val1, val2, label=label)
            del self._edge_refcounts[edge]

    def node_in(self, val):
        return val in self._node_refcounts

    def edge_in(self, val1, val2, label=None):
        return (val1, val2, label) in self._edge_refcounts

    def node_count(self, node):
        if node in self._node_refcounts:
            return self._node_refcounts[node]
        else:
            return 0

    def edge_count(self, val1, val2, label=None):
        edge = (val1, val2, label)
        if edge in self._edge_refcounts:
            return self._edge_refcounts[edge]
        else:
            return 0

    def __len__(self):
        return (reduce(lambda x, y: x+y, self._node_refcounts.values(), 0) +
                reduce(lambda x, y: x+y, self._edge_refcounts.values(), 0))

    def __str__(self):
        s = "{"
        for node in self.nodes:
            s += "(%s *%s: [" % (str(node), self._node_refcounts[node])
            if node in self.edges:
                s += ", ".join(
                    ["%s *%d" %
                        (str(x), self.edge_count(node, x.node, x.label))
                        for x in self.edges[node]])
            s += "],\n"
        s += "}"
        return s


class OrderedSet(collections.MutableSet):
    """Provide sequence capabilities with rapid membership checks.

    Mostly lifted from the activestate recipe[1] linked at Python's collections
    documentation[2]. Some modifications, such as returning True or False from
    add(key) and discard(key) if a change is made.

    [1] - http://code.activestate.com/recipes/576694/
    [2] - https://docs.python.org/2/library/collections.html
    """
    def __init__(self, iterable=None):
        self.end = end = []
        end += [None, end, end]         # sentinel node for doubly linked list
        self.map = {}                   # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]
            return True
        return False

    def discard(self, key):
        if key in self.map:
            key, prev, next = self.map.pop(key)
            prev[2] = next
            next[1] = prev
            return True
        return False

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        if not self:
            raise KeyError('pop from an empty set')
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        else:
            return False


class iterstr(object):
    """Lazily provides informal string representation of iterables.

    Calling __str__ directly on iterables returns a string containing the
    formal representation of the elements. This class wraps the iterable and
    instead returns the informal representation of the elements.
    """

    def __init__(self, iterable):
        self.iterable = iterable
        self._str_interp = None
        self._repr_interp = None

    def __str__(self):
        if self._str_interp is None:
            self._str_interp = "[" + ";".join(map(str, self.iterable)) + "]"
        return self._str_interp

    def __repr__(self):
        if self._repr_interp is None:
            self._repr_interp = "[" + ";".join(map(repr, self.iterable)) + "]"
        return self._repr_interp



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import functools
import logging

import eventlet
from mox3 import mox
from six.moves import range

from congress.datalog import compile
from congress.dse import dataobj
from congress import harness
from congress.policy_engines import agnostic
from congress.tests import base
from congress.tests import helper


LOG = logging.getLogger(__name__)


class BenchmarkDatasource(base.Benchmark):

    def setUp(self):
        super(BenchmarkDatasource, self).setUp()
        config = {'benchmark': {
                  'module': helper.data_module_path('benchmark_driver.py'),
                  'poll_time': 0}}
        cage = harness.create(helper.root_path(), None, config)
        engine = cage.service_object('engine')
        api = {'policy': cage.service_object('api-policy'),
               'rule': cage.service_object('api-rule'),
               'table': cage.service_object('api-table'),
               'row': cage.service_object('api-row'),
               'datasource': cage.service_object('api-datasource'),
               'status': cage.service_object('api-status'),
               'schema': cage.service_object('api-schema')}
        helper.retry_check_subscriptions(engine, [(api['rule'].name,
                                         'policy-update')])
        helper.retry_check_subscribers(api['rule'], [(engine.name,
                                       'policy-update')])
        self.assertTrue('benchmark' in cage.services)
        datasource = cage.service_object('benchmark')
        table_name = datasource.BENCHTABLE

        self.assertEqual(datasource.state, {})

        # add a subscriber to ensure the updates end up in datasource.dataPath
        pubdata = datasource.pubdata.setdefault(table_name,
                                                dataobj.pubData(table_name))
        pubdata.addsubscriber(self.__class__.__name__, "push", "")
        self.assertTrue(datasource.pubdata[table_name])

        self.cage = cage
        self.engine = engine
        self.api = api
        self.table_name = table_name
        self.datasource = datasource

    def benchmark_datasource_update(self, size):
        """Benchmark a datasource update.

        Time the propagation of a datasource update from datasource.poll() to
        ending up in the datasource.dataPath queue.
        """

        LOG.info("%s:: benchmarking datasource update of %d rows", size)
        self.datasource.datarows = size

        # intercept the queue addition so it doesn't immediately get pulled off
        # by the d6cage
        received = eventlet.Queue()
        self.mox.StubOutWithMock(self.datasource.dataPath, "put_nowait")
        self.datasource.dataPath.put_nowait(mox.IgnoreArg()).WithSideEffects(
            received.put_nowait)
        self.mox.ReplayAll()

        # poll and then wait until we've got an item from our queue
        LOG.info("%s:: polling datasource", self.__class__.__name__)
        self.datasource.poll()
        result = received.get(timeout=30)
        self.assertTrue(result.body)
        self.assertEqual(len(result.body.data), size)
        self.mox.VerifyAll()

    def benchmark_datasource_to_policy_update(self, size):
        """Benchmark small datsource update to policy propagation.

        Time the propagation of a datasource update from datasource.poll() to
        completion of a simple policy update.
        """
        LOG.info("%s:: benchmarking datasource update of %d rows", size)
        self.datasource.datarows = size
        table_name = self.table_name

        # dummy policy only intended to produce a subscriber for the table
        key_to_index = self.datasource.get_column_map(table_name)
        id_index = 'x%d' % list(key_to_index.items())[0][1]
        max_index = max(key_to_index.values())
        args = ['x%d' % i for i in range(max_index + 1)]
        formula = compile.parse1('p(%s) :- benchmark:%s(%s)' % (id_index,
                                 table_name, ','.join(args)))

        # publish the formula and verify we see a subscription
        LOG.debug('%s:: sending formula: %s', self.__class__.__name__, formula)
        self.api['rule'].publish('policy-update', [agnostic.Event(formula)])
        helper.retry_check_subscriptions(
            self.engine, [('benchmark', table_name)])
        helper.retry_check_subscribers(
            self.datasource, [(self.engine.name, table_name)])

        # intercept inbox.task_done() so we know when it's finished. Sadly,
        # eventlet doesn't have a condition-like object.
        fake_condition = eventlet.Queue()
        fake_notify = functools.partial(fake_condition.put_nowait, True)
        self.mox.StubOutWithMock(self.engine.inbox, "task_done")
        self.engine.inbox.task_done().WithSideEffects(fake_notify)
        self.mox.ReplayAll()

        LOG.info("%s:: polling datasource", self.__class__.__name__)
        self.datasource.poll()
        fake_condition.get(timeout=30)
        self.mox.VerifyAll()

    def test_benchmark_datasource_update_small(self):
        """Benchmark a small datasource update.

        Time the propagation of a small (10 row) datasource update from
        datasource.poll() to ending up in the datasource.dataPath queue.
        """
        self.benchmark_datasource_update(10)

    def test_benchmark_datasource_update_large(self):
        """Benchmark a large datasource update.

        Time the propagation of a large (100k row) datasource update from
        datasource.poll() to ending up in the datasource.dataPath queue.
        """
        self.benchmark_datasource_update(100000)

    def test_benchmark_datasource_to_policy_update_small(self):
        """Benchmark small datsource update to policy propagation.

        Time the propagation of a small (10 row) datasource update from
        datasource.poll() to a simple policy update.
        """
        self.benchmark_datasource_to_policy_update(10)

    def test_benchmark_datasource_to_policy_update_large(self):
        """Benchmark small datsource update to policy propagation.

        Time the propagation of a large (100k row) datasource update from
        datasource.poll() to a simple policy update.
        """
        self.benchmark_datasource_to_policy_update(100000)



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""RequestContext: context for requests that persist through congress."""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import copy
import datetime

from oslo_context import context as common_context
from oslo_log import log as logging

from congress.common import policy


LOG = logging.getLogger(__name__)


class RequestContext(common_context.RequestContext):
    """Security context and request information.

    Represents the user taking a given action within the system.

    """

    def __init__(self, user_id, tenant_id, is_admin=None, read_deleted="no",
                 roles=None, timestamp=None, load_admin_roles=True,
                 request_id=None, tenant_name=None, user_name=None,
                 overwrite=True, **kwargs):
        """Object initialization.

        :param read_deleted: 'no' indicates deleted records are hidden, 'yes'
            indicates deleted records are visible, 'only' indicates that
            *only* deleted records are visible.

        :param overwrite: Set to False to ensure that the greenthread local
            copy of the index is not overwritten.

        :param kwargs: Extra arguments that might be present, but we ignore
            because they possibly came in from older rpc messages.
        """
        super(RequestContext, self).__init__(user=user_id, tenant=tenant_id,
                                             is_admin=is_admin,
                                             request_id=request_id,
                                             overwrite=overwrite)
        self.user_name = user_name
        self.tenant_name = tenant_name

        self.read_deleted = read_deleted
        if not timestamp:
            timestamp = datetime.datetime.utcnow()
        self.timestamp = timestamp
        self._session = None
        self.roles = roles or []
        if self.is_admin is None:
            self.is_admin = policy.check_is_admin(self)

        # Log only once the context has been configured to prevent
        # format errors.
        if kwargs:
            LOG.debug(_('Arguments dropped when creating '
                        'context: %s'), kwargs)

    @property
    def project_id(self):
        return self.tenant

    @property
    def tenant_id(self):
        return self.tenant

    @tenant_id.setter
    def tenant_id(self, tenant_id):
        self.tenant = tenant_id

    @property
    def user_id(self):
        return self.user

    @user_id.setter
    def user_id(self, user_id):
        self.user = user_id

    def _get_read_deleted(self):
        return self._read_deleted

    def _set_read_deleted(self, read_deleted):
        if read_deleted not in ('no', 'yes', 'only'):
            raise ValueError(_("read_deleted can only be one of 'no', "
                               "'yes' or 'only', not %r") % read_deleted)
        self._read_deleted = read_deleted

    def _del_read_deleted(self):
        del self._read_deleted

    read_deleted = property(_get_read_deleted, _set_read_deleted,
                            _del_read_deleted)

    def to_dict(self):
        ret = super(RequestContext, self).to_dict()
        ret.update({'user_id': self.user_id,
                    'tenant_id': self.tenant_id,
                    'project_id': self.project_id,
                    'read_deleted': self.read_deleted,
                    'roles': self.roles,
                    'timestamp': str(self.timestamp),
                    'tenant_name': self.tenant_name,
                    'project_name': self.tenant_name,
                    'user_name': self.user_name})
        return ret

    @classmethod
    def from_dict(cls, values):
        return cls(**values)

    def elevated(self, read_deleted=None):
        """Return a version of this context with admin flag set."""
        context = copy.copy(self)
        context.is_admin = True

        if 'admin' not in [x.lower() for x in context.roles]:
            context.roles.append('admin')

        if read_deleted is not None:
            context.read_deleted = read_deleted

        return context


def get_admin_context(read_deleted="no", load_admin_roles=True):
    return RequestContext(user_id=None,
                          tenant_id=None,
                          is_admin=True,
                          read_deleted=read_deleted,
                          load_admin_roles=load_admin_roles,
                          overwrite=False)



# Copyright 2012 OpenStack Foundation
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2010 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import errno
import re
import socket
import ssl
import sys

import eventlet
import eventlet.wsgi
import greenlet
from oslo_log import log as logging
from oslo_service import service


LOG = logging.getLogger(__name__)


class EventletFilteringLogger(object):
    # NOTE(morganfainberg): This logger is designed to filter out specific
    # Tracebacks to limit the amount of data that eventlet can log. In the
    # case of broken sockets (EPIPE and ECONNRESET), we are seeing a huge
    # volume of data being written to the logs due to ~14 lines+ per traceback.
    # The traceback in these cases are, at best, useful for limited debugging
    # cases.
    def __init__(self, logger):
        self.logger = logger
        self.level = logger.logger.level
        self.regex = re.compile(r'errno (%d|%d)' %
                                (errno.EPIPE, errno.ECONNRESET), re.IGNORECASE)

    def write(self, msg):
        m = self.regex.search(msg)
        if m:
            self.logger.log(logging.logging.DEBUG,
                            'Error(%s) writing to socket.',
                            m.group(1))
        else:
            self.logger.log(self.level, msg.rstrip())


class Server(service.ServiceBase):
    """Server class to manage multiple WSGI sockets and applications."""

    def __init__(self, application, host=None, port=None, threads=1000,
                 keepalive=False, keepidle=None):
        self.application = application
        self.host = host or '0.0.0.0'
        self.port = port or 0
        self.pool = eventlet.GreenPool(threads)
        self.socket_info = {}
        self.greenthread = None
        self.do_ssl = False
        self.cert_required = False
        self.keepalive = keepalive
        self.keepidle = keepidle
        self.socket = None

    def start(self, key=None, backlog=128):
        """Run a WSGI server with the given application."""

        if self.socket is None:
            self.listen(key=key, backlog=backlog)

        self.greenthread = self.pool.spawn(self._run,
                                           self.application,
                                           self.socket)

    def listen(self, key=None, backlog=128):
        """Create and start listening on socket.

        Call before forking worker processes.

        Raises Exception if this has already been called.
        """

        if self.socket is not None:
            raise Exception(_('Server can only listen once.'))

        LOG.info(_('Starting %(arg0)s on %(host)s:%(port)s'),
                 {'arg0': sys.argv[0],
                  'host': self.host,
                  'port': self.port})

        # TODO(dims): eventlet's green dns/socket module does not actually
        # support IPv6 in getaddrinfo(). We need to get around this in the
        # future or monitor upstream for a fix
        info = socket.getaddrinfo(self.host,
                                  self.port,
                                  socket.AF_UNSPEC,
                                  socket.SOCK_STREAM)[0]
        _socket = eventlet.listen(info[-1],
                                  family=info[0],
                                  backlog=backlog)
        if key:
            self.socket_info[key] = _socket.getsockname()
        # SSL is enabled
        if self.do_ssl:
            if self.cert_required:
                cert_reqs = ssl.CERT_REQUIRED
            else:
                cert_reqs = ssl.CERT_NONE
            sslsocket = eventlet.wrap_ssl(_socket, certfile=self.certfile,
                                          keyfile=self.keyfile,
                                          server_side=True,
                                          cert_reqs=cert_reqs,
                                          ca_certs=self.ca_certs)
            _socket = sslsocket

        # Optionally enable keepalive on the wsgi socket.
        if self.keepalive:
            _socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

            # This option isn't available in the OS X version of eventlet
            if hasattr(socket, 'TCP_KEEPIDLE') and self.keepidle is not None:
                _socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE,
                                   self.keepidle)

        self.socket = _socket

    def set_ssl(self, certfile, keyfile=None, ca_certs=None,
                cert_required=True):
        self.certfile = certfile
        self.keyfile = keyfile
        self.ca_certs = ca_certs
        self.cert_required = cert_required
        self.do_ssl = True

    def kill(self):
        if self.greenthread is not None:
            self.greenthread.kill()

    def stop(self):
        self.kill()

    def reset(self):
        LOG.info("reset() not implemented yet")

    def wait(self):
        """Wait until all servers have completed running."""
        try:
            self.pool.waitall()
        except KeyboardInterrupt:
            pass
        except greenlet.GreenletExit:
            pass

    def _run(self, application, socket):
        """Start a WSGI server in a new green thread."""
        logger = logging.getLogger('eventlet.wsgi.server')
        try:
            eventlet.wsgi.server(socket, application, max_size=1000,
                                 log=EventletFilteringLogger(logger),
                                 debug=False)
        except greenlet.GreenletExit:
            # Wait until all servers have completed running
            pass
        except Exception:
            LOG.exception(_('Server error'))
            raise



# Copyright (c) 2016 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import eventlet

from oslo_config import cfg
from oslo_messaging import conffixture

from congress.dse2.data_service import DataService
from congress.tests import base
from congress.tests import helper


# Leave this in place for manual testing.
# For manual testing, support using rabbit driver instead of fake
USE_RABBIT = False
# if len(sys.argv) > 1:
#     driver_flg = sys.argv[1].lower()
#     if driver_flg == '--rabbit':
#         USE_RABBIT = True
#     elif driver_flg != '--fake':
#         print("Usage: %s [--fake | --rabbit]" % sys.argv[0])
#         sys.exit(1)
#     sys.argv[1:] = sys.argv[2:]


class _PingRpcEndpoint(object):
    def __init__(self, node_id):
        self.node_id = node_id
        self.ping_receive_count = 0
        self.ping_received_from = []

    def ping(self, client_ctxt, **args):
        self.ping_receive_count += 1
        self.ping_received_from.append(client_ctxt)
        return args


class _PingRpcService(DataService):
    def __init__(self, service_id, node_id):
        self.endpoints = [_PingRpcEndpoint(node_id)]
        super(_PingRpcService, self).__init__(service_id)

    def rpc_endpoints(self):
        return self.endpoints


class TestDseNode(base.TestCase):

    def setUp(self):
        super(TestDseNode, self).setUp()

        if USE_RABBIT:
            self.messaging_config = cfg.CONF
        else:
            mc_fixture = conffixture.ConfFixture(cfg.CONF)
            mc_fixture.conf.transport_url = 'kombu+memory://'
            self.messaging_config = mc_fixture.conf
        self.messaging_config.rpc_response_timeout = 1

    def test_start_stop(self):
        # create node and register services
        node = helper.make_dsenode_new_partition('test_node',
                                                 self.messaging_config, [])
        services = []
        for i in range(2):
            service = DataService('test-service-%s' % i)
            node.register_service(service)
            services.append(service)
        for s in node.get_services(True):
            self.assertTrue(s._running,
                            "Service '%s' started" % str(s))
        self.assertEqual(set(services), set(node.get_services()),
                         "All services accounted for on node.")
        self.assertTrue(node._rpcserver._started,
                        "RPC server is started")
        self.assertTrue(node._control_bus._running,
                        "Control Bus is started")

        # stop node
        node.stop()
        node.wait()
        self.assertFalse(node._running,
                         "Node is stopped after node start")
        for idx, s in enumerate(node.get_services(True)):
            self.assertFalse(s._running,
                             "Service '%s' stopped after node stop" % str(s))
        # TODO(pballand): fix bug
        # self.assertFalse(node._rpcserver._started,
        #                  "RPC server is stopped after node stop")
        self.assertFalse(node._control_bus._running,
                         "Control Bus is stopped after node stop")

        # restart node
        node.start()
        for s in node.get_services(True):
            self.assertTrue(s._running,
                            "Service '%s' started" % str(s))
        self.assertEqual(set(services), set(node.get_services()),
                         "All services accounted for on node.")
        self.assertTrue(node._rpcserver._started,
                        "RPC server is started")
        self.assertTrue(node._control_bus._running,
                        "Control Bus is started")

    def test_context(self):
        # Context must not only rely on node_id to prohibit multiple instances
        # of a node_id on the DSE
        part = helper.get_new_partition()
        n1 = helper.make_dsenode_same_partition(part, 'node_id',
                                                self.messaging_config, [])
        n2 = helper.make_dsenode_same_partition(part, 'node_id',
                                                self.messaging_config, [])
        self.assertEqual(n1._message_context, n1._message_context,
                         "Comparison of context from the same node is equal")
        self.assertNotEqual(n1._message_context, n2._message_context,
                            "Comparison of context from the different nodes "
                            "is not equal")

    def test_node_rpc(self):
        """Validate calling RPCs on DseNode"""
        part = helper.get_new_partition()
        nodes = []
        endpoints = []
        for i in range(3):
            nid = 'rpcnode%s' % i
            endpoints.append(_PingRpcEndpoint(nid))
            nodes.append(
                helper.make_dsenode_same_partition(
                    part, nid, self.messaging_config, [endpoints[-1]]))

        # Send from each node to each other node
        for i, source in enumerate(nodes):
            # intentionally including self in RPC target
            for j, target in enumerate(nodes):
                scount = endpoints[j].ping_receive_count
                args = {'arg1': 1, 'arg2': 'a'}
                ret = source.invoke_node_rpc(target.node_id, 'ping', **args)
                self.assertEqual(ret, args, "Ping echoed arguments")
                ecount = endpoints[j].ping_receive_count
                self.assertEqual(ecount - scount, 1,
                                 "Node %s received ping (%s was sending)"
                                 % (nodes[j].node_id, nodes[i].node_id))
                self.assertEqual(
                    endpoints[j].ping_received_from[-1]['node_id'],
                    nodes[i].node_id,
                    "Last ping received on %s was from %s" % (
                        nodes[j].node_id, nodes[i].node_id))

    def test_node_broadcast_rpc(self):
        """Validate calling RPCs on DseNode"""
        part = helper.get_new_partition()
        nodes = []
        endpoints = []
        for i in range(3):
            nid = 'rpcnode%s' % i
            endpoints.append(_PingRpcEndpoint(nid))
            nodes.append(
                helper.make_dsenode_same_partition(
                    part, nid, self.messaging_config, [endpoints[-1]]))

        # Send from each node to all other nodes
        for i, source in enumerate(nodes):
            scounts = []
            for j, target in enumerate(nodes):
                scounts.append(endpoints[j].ping_receive_count)
            source.broadcast_node_rpc('ping', arg1=1, arg2='a')
            eventlet.sleep(0.5)  # wait for async delivery
            for j, target in enumerate(nodes):
                ecount = endpoints[j].ping_receive_count
                self.assertEqual(ecount - scounts[j], 1,
                                 "Node %s received ping (%s was sending)"
                                 % (nodes[j].node_id, source.node_id))
                self.assertEqual(
                    endpoints[j].ping_received_from[-1]['node_id'],
                    source.node_id,
                    "Last ping received on %s was from %s" % (
                        nodes[j].node_id, source.node_id))

    def test_service_rpc(self):
        part = helper.get_new_partition()
        nodes = []
        services = []
        for i in range(3):
            nid = 'svc_rpc_node%s' % i
            node = helper.make_dsenode_same_partition(
                part, nid, self.messaging_config)
            service = _PingRpcService('srpc_node_svc%s' % i, nid)
            node.register_service(service)
            nodes.append(node)
            services.append(service)

        # Send from each node to each other node
        for i, source in enumerate(nodes):
            # intentionally including self in RPC target
            for j, service in enumerate(services):
                ep = nodes[j]._services[-1].endpoints[0]
                scount = ep.ping_receive_count
                args = {'arg1': 1, 'arg2': 'a'}
                ret = source.invoke_service_rpc(service.service_id, 'ping',
                                                **args)
                self.assertEqual(ret, args, "Ping echoed arguments")
                ecount = ep.ping_receive_count
                self.assertEqual(ecount - scount, 1,
                                 "Node %s received ping (%s was sending)"
                                 % (nodes[j].node_id, nodes[i].node_id))
                self.assertEqual(
                    ep.ping_received_from[-1]['node_id'],
                    nodes[i].node_id,
                    "Last ping received on %s was from %s" % (
                        nodes[j].node_id, nodes[i].node_id))

    def test_broadcast_service_rpc(self):
        part = helper.get_new_partition()
        nodes = []
        services = []
        for i in range(3):
            nid = 'svc_rpc_node%s' % i
            node = helper.make_dsenode_same_partition(
                part, nid, self.messaging_config)
            service = _PingRpcService('tbsr_svc', nid)
            node.register_service(service)
            nodes.append(node)
            services.append(service)

        # Send from each node to all services
        for i, source in enumerate(nodes):
            scounts = []
            for j, target in enumerate(nodes):
                ep = nodes[j]._services[-1].endpoints[0]
                scounts.append(ep.ping_receive_count)
            source.broadcast_service_rpc('tbsr_svc', 'ping', arg1=1, arg2='a')
            eventlet.sleep(0.5)  # wait for async delivery
            for j, target in enumerate(nodes):
                ep = nodes[j]._services[-1].endpoints[0]
                ecount = ep.ping_receive_count
                self.assertEqual(ecount - scounts[j], 1,
                                 "Node %s received ping (%s was sending)"
                                 % (nodes[j].node_id, source.node_id))
                self.assertEqual(
                    ep.ping_received_from[-1]['node_id'],
                    source.node_id,
                    "Last ping received on %s was from %s" % (
                        nodes[j].node_id, source.node_id))

    def test_get_global_service_names(self):
        node = helper.make_dsenode_new_partition('test_node',
                                                 self.messaging_config, [])
        test1 = _PingRpcService('test1', 'test1')
        test2 = _PingRpcService('test2', 'test2')
        node.register_service(test1)
        node.register_service(test2)
        actual = set(node.get_global_service_names())
        self.assertEqual(actual, set(['test1', 'test2']))


# Leave this to make manual testing with RabbitMQ easy
# if __name__ == '__main__':
#     import unittest
#     unittest.main(verbosity=2)



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

try:
    # For Python 3
    import http.client as httplib
except ImportError:
    import httplib
import json
import uuid

import mock
import webob

from congress.api import webservice
from congress.tests import base


class TestSimpleDataModel(base.TestCase):
    # if random ID matches, go to Vegas or file a uuid library bug
    UNADDED_ID = str(uuid.uuid4())
    CONTEXTS = [None, {'a': 'ctxt1'}, {'b': 'ctxt2', 'c': 'ctxt3'}]

    def test_get_items(self):
        """Test API DataModel get_items functionality."""
        model = webservice.SimpleDataModel("test")
        for context in self.CONTEXTS:
            ret = model.get_items({}, context=context)
            self.assertEqual(
                ret.keys(), {'results': None}.keys(),
                "get_items returns dict with single 'results' key")
            self.assertEqual(
                tuple(ret['results']), tuple(),
                "get_items of empty model returns empty results list")
            items = [{"i1": "%s/foo" % context}, {"i2": "%s/bar" % context}]
            for item in items:
                model.add_item(item, {}, context=context)
            ret2 = model.get_items({}, context=context)
            self.assertEqual(sorted(list(ret2['results']),
                                    key=(lambda x: str(type(x)) + repr(x))),
                             sorted(items,
                                    key=(lambda x: str(type(x)) + repr(x))),
                             "get_items() returns all items added to model")

    def test_add_item(self):
        """Test API DataModel add_item functionality."""
        model = webservice.SimpleDataModel("test")
        assigned_ids = set()
        for context in self.CONTEXTS:
            items = ["%s/foo" % context, "%s/bar" % context]
            ret = model.add_item(items[0], {}, context=context)
            self.assertIsInstance(ret, tuple, "add_item returns a tuple")
            self.assertEqual(len(ret), 2,
                             "add_item returns tuple of length 2")
            self.assertNotIn(ret[0], assigned_ids,
                             "add_item assigned unique ID")
            assigned_ids.add(ret[0])
            self.assertEqual(ret[1], items[0], "add_item returned added item")

            ret = model.add_item(items[1], {}, 'myid', context=context)
            self.assertEqual(ret[0], 'myid',
                             "add_item returned provided ID")
            self.assertEqual(ret[1], items[1], "add_item returned added item")

    def test_get_item(self):
        """Test API DataModel get_item functionality."""
        model = webservice.SimpleDataModel("test")
        for context in self.CONTEXTS:
            items = ["%s/foo" % context, "%s/bar" % context]
            id_ = model.add_item(items[0], {}, context=context)[0]
            ret = model.get_item(id_, {}, context=context)
            self.assertEqual(ret, items[0],
                             "get_item(assigned_id) returns proper item")

            id_ = 'myid'
            ret = model.get_item(id_, {}, context=context)
            self.assertIsNone(ret,
                              "get_item(unadded_provided_id) returns None")
            model.add_item(items[1], {}, id_, context=context)
            ret = model.get_item(id_, {}, context=context)
            self.assertEqual(ret, items[1],
                             "get_item(provided_id) returned added item")

            ret = model.get_item(self.UNADDED_ID, {}, context=context)
            self.assertIsNone(ret, "get_item(unadded_id) returns None")

    def test_update_item(self):
        """Test API DataModel update_item functionality."""
        model = webservice.SimpleDataModel("test")
        for context in self.CONTEXTS:
            items = ["%s/foo%d" % (context, i) for i in [0, 1, 2]]
            id_, item = model.add_item(items[0], {}, context=context)
            self.assertNotEqual(item, items[1], "item not yet updated")
            ret = model.update_item(id_, items[1], {}, context=context)
            self.assertEqual(ret, items[1],
                             "update_item returned updated item")
            ret = model.get_item(id_, {}, context=context)
            self.assertEqual(ret, items[1],
                             "get_item(updated_item_id) returns updated item")

            self.assertNotEqual(item, items[2], "item not yet reupdated")
            ret = model.update_item(id_, items[2], {}, context=context)
            self.assertEqual(ret, items[2],
                             "update_item returned reupdated item")
            ret = model.get_item(id_, {}, context=context)
            self.assertEqual(
                ret, items[2],
                "get_item(reupdated_item_id) returns reupdated item")

            self.assertRaises(KeyError, model.update_item,
                              self.UNADDED_ID, 'blah', {}, context)

    def test_delete_item(self):
        """Test API DataModel delete_item functionality."""
        model = webservice.SimpleDataModel("test")

        for context in self.CONTEXTS:
            item_ids = []
            items = ["%s/foo%d" % (context, i) for i in [0, 1, 2]]
            for i in range(len(items)):
                id_, item = model.add_item(items[i], {}, context=context)
                item_ids.append(id_)

            for i in range(len(items)):
                ret = model.delete_item(item_ids[i], {}, context=context)
                self.assertEqual(ret, items[i],
                                 "delete_item returned deleted item")
                self.assertRaises(KeyError, model.delete_item, item_ids[i],
                                  {}, context)
            self.assertEqual(
                len(model.get_items({}, context=context)['results']), 0,
                "all items deleted")

            self.assertRaises(KeyError, model.delete_item, self.UNADDED_ID,
                              {}, context)


class TestAbstractHandler(base.TestCase):
    def test_parse_json_body(self):
        abstract_handler = webservice.AbstractApiHandler(r'/')
        request = mock.MagicMock()
        data = {"some": ["simple", "json"]}
        serialized_data = json.dumps({"some": ["simple", "json"]})
        invalid_json = 'this is not valid JSON'

        # correctly assume application/json when no content-type header
        request = webob.Request.blank('/')
        self.assertEqual(request.content_type, '')
        request.body = serialized_data.encode('utf-8')
        ret = abstract_handler._parse_json_body(request)
        self.assertEqual(ret, data)

        # correctly validate valid content-type headers
        for ct in ['application/json',
                   'Application/jSoN',
                   'application/json; charset=utf-8',
                   'apPLICAtion/JSOn; charset=UtF-8',
                   'apPLICAtion/JSOn; CHARset=utf-8; IGnored=c',
                   'application/json; ignored_param=a; ignored2=b']:
            request = webob.Request.blank('/', content_type=ct)
            request.body = serialized_data.encode('utf-8')
            try:
                ret = abstract_handler._parse_json_body(request)
            except Exception:
                self.fail("accepts content type '%s'" % ct)
            self.assertEqual(ret, data, "Accepts content type '%s'" % ct)

        # correctly fail on invalid content-type headers
        request = webob.Request.blank('/', content_type='text/json')
        request.body = serialized_data.encode('utf-8')
        self.assertRaises(webservice.DataModelException,
                          abstract_handler._parse_json_body, request)

        # enforce unspecified or utf-8 charset
        # valid charset checked above, just need to check invalid
        request = webob.Request.blank(
            '/', content_type='application/json; charset=utf-16')
        request.body = serialized_data.encode('utf-8')
        self.assertRaises(webservice.DataModelException,
                          abstract_handler._parse_json_body, request)

        # raise DataModelException on non-JSON body
        request = webob.Request.blank(
            '/', content_type='application/json; charset=utf-8')
        request.body = invalid_json.encode('utf-8')
        self.assertRaises(webservice.DataModelException,
                          abstract_handler._parse_json_body, request)


class TestElementHandler(base.TestCase):
    def test_read(self):
        # TODO(pballand): write tests
        pass

    def test_action(self):
        element_handler = webservice.ElementHandler(r'/', '')
        element_handler.model = webservice.SimpleDataModel("test")
        request = mock.MagicMock()
        request.path = "/"

        response = element_handler.action(request)
        self.assertEqual(400, response.status_code)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(json.loads(
                         response.body.decode('utf-8'))['error']['message'],
                         "Missing required action parameter.")

        request.params = mock.MagicMock()
        request.params.getall.return_value = ['do_test']
        request.params["action"] = "do_test"
        request.path = "/"
        response = element_handler.action(request)
        self.assertEqual(501, response.status_code)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(json.loads(
                         response.body.decode('utf-8'))['error']['message'],
                         "Method not supported")

        # test action impl returning python primitives
        simple_data = [1, 2]
        element_handler.model.do_test_action = lambda *a, **kwa: simple_data
        response = element_handler.action(request)
        self.assertEqual(200, response.status_code)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(json.loads(response.body.decode('utf-8')),
                         simple_data)

        # test action impl returning custom webob response
        custom_data = webob.Response(body="test".encode('utf-8'), status=599,
                                     content_type="custom/test")
        element_handler.model.do_test_action = lambda *a, **kwa: custom_data
        response = element_handler.action(request)
        self.assertEqual(599, response.status_code)
        self.assertEqual('custom/test', response.content_type)
        self.assertEqual(response.body.decode('utf-8'), "test")

    def test_replace(self):
        # TODO(pballand): write tests
        pass

    def test_update(self):
        # TODO(pballand): write tests
        pass

    def test_delete(self):
        # TODO(pballand): write tests
        pass


class TestCollectionHandler(base.TestCase):
    def test_get_action_type(self):
        collection_handler = webservice.CollectionHandler(r'/', '')
        self.assertEqual('get',
                         collection_handler._get_action_type("GET"))
        self.assertEqual('create',
                         collection_handler._get_action_type("POST"))
        self.assertEqual('delete',
                         collection_handler._get_action_type("DELETE"))
        self.assertEqual('update',
                         collection_handler._get_action_type("PATCH"))
        self.assertEqual('update',
                         collection_handler._get_action_type("PUT"))
        self.assertRaises(TypeError, collection_handler._get_action_type,
                          'Wah!')

    def test_create_member(self):
        collection_handler = webservice.CollectionHandler(r'/', '')
        collection_handler.model = webservice.SimpleDataModel("test")
        request = webob.Request.blank('/')
        request.content_type = 'application/json'
        request.body = '{"key": "value"}'.encode('utf-8')
        response = collection_handler.create_member(request, id_='123')
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(str(httplib.CREATED) + " Created", response.status)
        self.assertEqual("%s/%s" % (request.path, '123'), response.location)
        actual_response = json.loads(response.body.decode('utf-8'))
        actual_id = actual_response.get("id")
        actual_value = actual_response.get("key")
        self.assertEqual('123', actual_id)
        self.assertEqual('value', actual_value)

    def test_list_members(self):
        collection_handler = webservice.CollectionHandler(r'/', '')
        collection_handler.model = webservice.SimpleDataModel("test")
        request = mock.MagicMock()
        request.body = '{"key": "value"}'
        request.params = mock.MagicMock()
        request.path = "/"
        response = collection_handler.list_members(request)
        items = collection_handler.model.get_items(
            request.params,
            context=collection_handler._get_context(request))

        expected_body = ("%s\n" % json.dumps(items, indent=2)).encode('utf-8')
        self.assertEqual('application/json', response.content_type)

        self.assertEqual(expected_body, response.body)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(str(httplib.OK) + " OK", response.status)

    def test_update_members(self):
        collection_handler = webservice.CollectionHandler(r'/', '')
        collection_handler.model = webservice.SimpleDataModel('test')
        request = webob.Request.blank('/')
        request.content_type = 'application/json'
        request.body = '{"key1": "value1", "key2": "value2"}'.encode('utf-8')
        response = collection_handler.update_members(request)

        self.assertEqual('application/json', response.content_type)
        self.assertEqual(str(httplib.OK) + " OK", response.status)
        expected_items = {
            "key1": "value1",
            "key2": "value2",
            }
        self.assertEqual(expected_items, collection_handler.model.items)



# Copyright (c) 2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_utils import uuidutils

from congress.db import datasources
from congress.tests import base


class TestDbDatasource(base.SqlTestCase):

    def test_add_datasource(self):
        id_ = uuidutils.generate_uuid()
        source = datasources.add_datasource(
            id_=id_,
            name="hiya",
            driver="foo",
            config='{user: foo}',
            description="hello",
            enabled=True)
        self.assertEqual(id_, source.id)
        self.assertEqual("hiya", source.name)
        self.assertEqual("foo", source.driver)
        self.assertEqual("hello", source.description)
        self.assertEqual('"{user: foo}"', source.config)
        self.assertTrue(source.enabled)

    def test_delete_datasource(self):
        id_ = uuidutils.generate_uuid()
        datasources.add_datasource(
            id_=id_,
            name="hiya",
            driver="foo",
            config='{user: foo}',
            description="hello",
            enabled=True)
        self.assertTrue(datasources.delete_datasource(id_))

    def test_delete_non_existing_datasource(self):
        self.assertFalse(datasources.delete_datasource('no_id'))

    def test_get_datasource_by_name(self):
        id_ = uuidutils.generate_uuid()
        datasources.add_datasource(
            id_=id_,
            name="hiya",
            driver="foo",
            config='{user: foo}',
            description="hello",
            enabled=True)
        source = datasources.get_datasource_by_name('hiya')
        self.assertEqual(id_, source.id)
        self.assertEqual("hiya", source.name)
        self.assertEqual("foo", source.driver)
        self.assertEqual("hello", source.description)
        self.assertEqual('"{user: foo}"', source.config)
        self.assertTrue(source.enabled)

    def test_get_datasource_by_id(self):
        id_ = uuidutils.generate_uuid()
        datasources.add_datasource(
            id_=id_,
            name="hiya",
            driver="foo",
            config='{user: foo}',
            description="hello",
            enabled=True)
        source = datasources.get_datasource(id_)
        self.assertEqual(id_, source.id)
        self.assertEqual("hiya", source.name)
        self.assertEqual("foo", source.driver)
        self.assertEqual("hello", source.description)
        self.assertEqual('"{user: foo}"', source.config)
        self.assertTrue(source.enabled)

    def test_get_datasource(self):
        id_ = uuidutils.generate_uuid()
        datasources.add_datasource(
            id_=id_,
            name="hiya",
            driver="foo",
            config='{user: foo}',
            description="hello",
            enabled=True)
        sources = datasources.get_datasources()
        self.assertEqual(id_, sources[0].id)
        self.assertEqual("hiya", sources[0].name)
        self.assertEqual("foo", sources[0].driver)
        self.assertEqual("hello", sources[0].description)
        self.assertEqual('"{user: foo}"', sources[0].config)
        self.assertTrue(sources[0].enabled)



#!/usr/bin/env python
# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
import neutronclient.v2_0.client
from oslo_log import log as logging

from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils

LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    """This method is called by d6cage to create a dataservice instance."""
    return NeutronV2Driver(name, keys, inbox, datapath, args)


class NeutronV2Driver(datasource_driver.PollingDataSourceDriver,
                      datasource_driver.ExecutionDriver):

    NETWORKS = 'networks'
    FIXED_IPS = 'fixed_ips'
    SECURITY_GROUP_PORT_BINDINGS = 'security_group_port_bindings'
    PORTS = 'ports'
    ALLOCATION_POOLS = 'allocation_pools'
    DNS_NAMESERVERS = 'dns_nameservers'
    HOST_ROUTES = 'host_routes'
    SUBNETS = 'subnets'
    EXTERNAL_FIXED_IPS = 'external_fixed_ips'
    EXTERNAL_GATEWAY_INFOS = 'external_gateway_infos'
    ROUTERS = 'routers'
    SECURITY_GROUP_RULES = 'security_group_rules'
    SECURITY_GROUPS = 'security_groups'
    FLOATING_IPS = 'floating_ips'

    # This is the most common per-value translator, so define it once here.
    value_trans = {'translation-type': 'VALUE'}

    floating_ips_translator = {
        'translation-type': 'HDICT',
        'table-name': FLOATING_IPS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'The UUID of the floating IP address',
              'translator': value_trans},
             {'fieldname': 'router_id', 'desc': 'UUID of router',
              'translator': value_trans},
             {'fieldname': 'tenant_id', 'desc': 'Tenant ID',
              'translator': value_trans},
             {'fieldname': 'floating_network_id',
              'desc': 'The UUID of the network associated with floating IP',
              'translator': value_trans},
             {'fieldname': 'fixed_ip_address',
              'desc': 'Fixed IP address associated with floating IP address',
              'translator': value_trans},
             {'fieldname': 'floating_ip_address',
              'desc': 'The floating IP address', 'translator': value_trans},
             {'fieldname': 'port_id', 'desc': 'UUID of port',
              'translator': value_trans},
             {'fieldname': 'status', 'desc': 'The floating IP status',
              'translator': value_trans})}

    networks_translator = {
        'translation-type': 'HDICT',
        'table-name': NETWORKS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'Network ID',
              'translator': value_trans},
             {'fieldname': 'tenant_id', 'desc': 'Tenant ID',
              'translator': value_trans},
             {'fieldname': 'name', 'desc': 'Network name',
              'translator': value_trans},
             {'fieldname': 'status', 'desc': 'Network status',
              'translator': value_trans},
             {'fieldname': 'admin_state_up',
              'desc': 'Administrative state of the network (true/false)',
              'translator': value_trans},
             {'fieldname': 'shared',
              'desc': 'Indicates if network is shared across all tenants',
              'translator': value_trans})}

    ports_fixed_ips_translator = {
        'translation-type': 'HDICT',
        'table-name': FIXED_IPS,
        'parent-key': 'id',
        'parent-col-name': 'port_id',
        'parent-key-desc': 'UUID of Port',
        'selector-type': 'DICT_SELECTOR',
        'in-list': True,
        'field-translators':
            ({'fieldname': 'ip_address',
              'desc': 'The IP addresses for the port',
              'translator': value_trans},
             {'fieldname': 'subnet_id',
              'desc': 'The UUID of the subnet to which the port is attached',
              'translator': value_trans})}

    ports_security_groups_translator = {
        'translation-type': 'LIST',
        'table-name': SECURITY_GROUP_PORT_BINDINGS,
        'parent-key': 'id',
        'parent-col-name': 'port_id',
        'parent-key-desc': 'UUID of port',
        'val-col': 'security_group_id',
        'val-col-desc': 'UUID of security group',
        'translator': value_trans}

    ports_translator = {
        'translation-type': 'HDICT',
        'table-name': PORTS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'UUID of port',
              'translator': value_trans},
             {'fieldname': 'tenant_id', 'desc': 'tenant ID',
              'translator': value_trans},
             {'fieldname': 'name', 'desc': 'port name',
              'translator': value_trans},
             {'fieldname': 'network_id', 'desc': 'UUID of attached network',
              'translator': value_trans},
             {'fieldname': 'mac_address', 'desc': 'MAC address of the port',
              'translator': value_trans},
             {'fieldname': 'admin_state_up',
              'desc': 'Administrative state of the port',
              'translator': value_trans},
             {'fieldname': 'status', 'desc': 'Port status',
              'translator': value_trans},
             {'fieldname': 'device_id',
              'desc': 'The UUID of the device that uses this port',
              'translator': value_trans},
             {'fieldname': 'device_owner',
              'desc': 'The UUID of the entity that uses this port',
              'translator': value_trans},
             {'fieldname': 'fixed_ips',
              'desc': 'The IP addresses for the port',
              'translator': ports_fixed_ips_translator},
             {'fieldname': 'security_groups',
              'translator': ports_security_groups_translator})}

    subnets_allocation_pools_translator = {
        'translation-type': 'HDICT',
        'table-name': ALLOCATION_POOLS,
        'parent-key': 'id',
        'parent-col-name': 'subnet_id',
        'parent-key-desc': 'UUID of subnet',
        'selector-type': 'DICT_SELECTOR',
        'in-list': True,
        'field-translators':
            ({'fieldname': 'start',
              'desc': 'The start address for the allocation pools',
              'translator': value_trans},
             {'fieldname': 'end',
              'desc': 'The end address for the allocation pools',
              'translator': value_trans})}

    subnets_dns_nameservers_translator = {
        'translation-type': 'LIST',
        'table-name': DNS_NAMESERVERS,
        'parent-key': 'id',
        'parent-col-name': 'subnet_id',
        'parent-key-desc': 'UUID of subnet',
        'val-col': 'dns_nameserver',
        'val-col-desc': 'The DNS server',
        'translator': value_trans}

    subnets_routes_translator = {
        'translation-type': 'HDICT',
        'table-name': HOST_ROUTES,
        'parent-key': 'id',
        'parent-col-name': 'subnet_id',
        'parent-key-desc': 'UUID of subnet',
        'selector-type': 'DICT_SELECTOR',
        'in-list': True,
        'field-translators':
            ({'fieldname': 'destination',
              'desc': 'The destination for static route',
              'translator': value_trans},
             {'fieldname': 'nexthop',
              'desc': 'The next hop for the destination',
              'translator': value_trans})}

    subnets_translator = {
        'translation-type': 'HDICT',
        'table-name': SUBNETS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'UUID of subnet',
              'translator': value_trans},
             {'fieldname': 'tenant_id', 'desc': 'tenant ID',
              'translator': value_trans},
             {'fieldname': 'name', 'desc': 'subnet name',
              'translator': value_trans},
             {'fieldname': 'network_id', 'desc': 'UUID of attached network',
              'translator': value_trans},
             {'fieldname': 'ip_version',
              'desc': 'The IP version, which is 4 or 6',
              'translator': value_trans},
             {'fieldname': 'cidr', 'desc': 'The CIDR',
              'translator': value_trans},
             {'fieldname': 'gateway_ip', 'desc': 'The gateway IP address',
              'translator': value_trans},
             {'fieldname': 'enable_dhcp', 'desc': 'Is DHCP is enabled or not',
              'translator': value_trans},
             {'fieldname': 'ipv6_ra_mode', 'desc': 'The IPv6 RA mode',
              'translator': value_trans},
             {'fieldname': 'ipv6_address_mode',
              'desc': 'The IPv6 address mode', 'translator': value_trans},
             {'fieldname': 'allocation_pools',
              'translator': subnets_allocation_pools_translator},
             {'fieldname': 'dns_nameservers',
              'translator': subnets_dns_nameservers_translator},
             {'fieldname': 'host_routes',
              'translator': subnets_routes_translator})}

    external_fixed_ips_translator = {
        'translation-type': 'HDICT',
        'table-name': EXTERNAL_FIXED_IPS,
        'parent-key': 'router_id',
        'parent-col-name': 'router_id',
        'parent-key-desc': 'UUID of router',
        'selector-type': 'DICT_SELECTOR',
        'in-list': True,
        'field-translators':
            ({'fieldname': 'subnet_id', 'desc': 'UUID of the subnet',
              'translator': value_trans},
             {'fieldname': 'ip_address', 'desc': 'IP Address',
              'translator': value_trans})}

    routers_external_gateway_infos_translator = {
        'translation-type': 'HDICT',
        'table-name': EXTERNAL_GATEWAY_INFOS,
        'parent-key': 'id',
        'parent-col-name': 'router_id',
        'parent-key-desc': 'UUID of router',
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'network_id', 'desc': 'Network ID',
              'translator': value_trans},
             {'fieldname': 'enable_snat',
              'desc': 'current Source NAT status for router',
              'translator': value_trans},
             {'fieldname': 'external_fixed_ips',
              'translator': external_fixed_ips_translator})}

    routers_translator = {
        'translation-type': 'HDICT',
        'table-name': ROUTERS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'uuid of the router',
              'translator': value_trans},
             {'fieldname': 'tenant_id', 'desc': 'tenant ID',
              'translator': value_trans},
             {'fieldname': 'status', 'desc': 'router status',
              'translator': value_trans},
             {'fieldname': 'admin_state_up',
              'desc': 'administrative state of router',
              'translator': value_trans},
             {'fieldname': 'name', 'desc': 'router name',
              'translator': value_trans},
             {'fieldname': 'distributed',
              'desc': "indicates if it's distributed router ",
              'translator': value_trans},
             {'fieldname': 'external_gateway_info',
              'translator': routers_external_gateway_infos_translator})}

    security_group_rules_translator = {
        'translation-type': 'HDICT',
        'table-name': SECURITY_GROUP_RULES,
        'parent-key': 'id',
        'parent-col-name': 'security_group_id',
        'parent-key-desc': 'uuid of security group',
        'selector-type': 'DICT_SELECTOR',
        'in-list': True,
        'field-translators':
            ({'fieldname': 'id', 'desc': 'The UUID of the security group rule',
              'translator': value_trans},
             {'fieldname': 'tenant_id', 'desc': 'tenant ID',
              'translator': value_trans},
             {'fieldname': 'remote_group_id',
              'desc': 'remote group id to associate with security group rule',
              'translator': value_trans},
             {'fieldname': 'direction',
              'desc': 'Direction in which the security group rule is applied',
              'translator': value_trans},
             {'fieldname': 'ethertype', 'desc': 'IPv4 or IPv6',
              'translator': value_trans},
             {'fieldname': 'protocol',
              'desc': 'protocol that is matched by the security group rule.',
              'translator': value_trans},
             {'fieldname': 'port_range_min',
              'desc': 'Min port number in the range',
              'translator': value_trans},
             {'fieldname': 'port_range_max',
              'desc': 'Max port number in the range',
              'translator': value_trans},
             {'fieldname': 'remote_ip_prefix',
              'desc': 'Remote IP prefix to be associated',
              'translator': value_trans})}

    security_group_translator = {
        'translation-type': 'HDICT',
        'table-name': SECURITY_GROUPS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'The UUID for the security group',
              'translator': value_trans},
             {'fieldname': 'tenant_id', 'desc': 'Tenant ID',
              'translator': value_trans},
             {'fieldname': 'name', 'desc': 'The security group name',
              'translator': value_trans},
             {'fieldname': 'description', 'desc': 'security group description',
              'translator': value_trans},
             {'fieldname': 'security_group_rules',
              'translator': security_group_rules_translator})}

    TRANSLATORS = [networks_translator, ports_translator, subnets_translator,
                   routers_translator, security_group_translator,
                   floating_ips_translator]

    def __init__(self, name='', keys='', inbox=None,
                 datapath=None, args=None):
        super(NeutronV2Driver, self).__init__(name, keys, inbox,
                                              datapath, args)
        datasource_driver.ExecutionDriver.__init__(self)
        self.creds = args
        self.neutron = neutronclient.v2_0.client.Client(**self.creds)
        self.add_executable_client_methods(self.neutron,
                                           'neutronclient.v2_0.client')
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'neutronv2'
        result['description'] = ('Datasource driver that interfaces with '
                                 'OpenStack Networking aka Neutron.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['secret'] = ['password']
        return result

    def update_from_datasource(self):

        LOG.debug("Neutron grabbing networks")
        networks = self.neutron.list_networks()
        self._translate_networks(networks)

        LOG.debug("Neutron grabbing ports")
        ports = self.neutron.list_ports()
        self._translate_ports(ports)

        subnets = self.neutron.list_subnets()
        self._translate_subnets(subnets)

        routers = self.neutron.list_routers()
        self._translate_routers(routers)

        security_groups = self.neutron.list_security_groups()
        self._translate_security_groups(security_groups)

        floating_ips = self.neutron.list_floatingips()
        self._translate_floating_ips(floating_ips)

    @ds_utils.update_state_on_changed(FLOATING_IPS)
    def _translate_floating_ips(self, obj):
        LOG.debug("floating_ips: %s", dict(obj))

        row_data = NeutronV2Driver.convert_objs(obj['floatingips'],
                                                self.floating_ips_translator)
        return row_data

    @ds_utils.update_state_on_changed(NETWORKS)
    def _translate_networks(self, obj):
        LOG.debug("networks: %s", dict(obj))

        row_data = NeutronV2Driver.convert_objs(obj['networks'],
                                                self.networks_translator)
        return row_data

    @ds_utils.update_state_on_changed(PORTS)
    def _translate_ports(self, obj):
        LOG.debug("ports: %s", obj)
        row_data = NeutronV2Driver.convert_objs(obj['ports'],
                                                self.ports_translator)
        return row_data

    @ds_utils.update_state_on_changed(SUBNETS)
    def _translate_subnets(self, obj):
        LOG.debug("subnets: %s", obj)
        row_data = NeutronV2Driver.convert_objs(obj['subnets'],
                                                self.subnets_translator)
        return row_data

    @ds_utils.update_state_on_changed(ROUTERS)
    def _translate_routers(self, obj):
        LOG.debug("routers: %s", obj)
        row_data = NeutronV2Driver.convert_objs(obj['routers'],
                                                self.routers_translator)
        return row_data

    @ds_utils.update_state_on_changed(SECURITY_GROUPS)
    def _translate_security_groups(self, obj):
        LOG.debug("security_groups: %s", obj)
        row_data = NeutronV2Driver.convert_objs(obj['security_groups'],
                                                self.security_group_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.neutron, action, action_args)



# Copyright (c) 2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from congress.api import error_codes
from congress.tests import base


class TestErrorCodes(base.TestCase):

    def setUp(self):
        super(TestErrorCodes, self).setUp()
        self.original_errors = error_codes.errors

    def tearDown(self):
        super(TestErrorCodes, self).tearDown()
        error_codes.errors = self.original_errors

    def test_get_error_code(self):
        name = 'fake-error'
        error_codes.errors = {
            "fake-error": (0000, 'This is a fake error code.', 400)
        }
        expected_num = 0000
        expected_desc = 'This is a fake error code.'
        expected_http = 400
        expected_ret = (expected_num, expected_desc)

        ret = error_codes.get(name)
        self.assertEqual(expected_ret, ret)

        num = error_codes.get_num(name)
        self.assertEqual(expected_num, num)

        desc = error_codes.get_desc(name)
        self.assertEqual(expected_desc, desc)

        http = error_codes.get_http(name)
        self.assertEqual(expected_http, http)

    def test_get_unknown__error_code(self):
        name = 'fake_error_code'
        error_codes.errors = {
            error_codes.UNKNOWN: (0000, 'Unknown error', 400),
            'fake-error': (1000, 'Fake error', 404)
        }
        expected_num = 0000
        expected_desc = 'Unknown error'
        expected_http = 400
        expected_ret = (expected_num, expected_desc)

        ret = error_codes.get(name)
        self.assertEqual(expected_ret, ret)

        num = error_codes.get_num(name)
        self.assertEqual(expected_num, num)

        desc = error_codes.get_desc(name)
        self.assertEqual(expected_desc, desc)

        http = error_codes.get_http(name)
        self.assertEqual(expected_http, http)



# Copyright (c) 2015 NTT, OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from congress.api import api_utils
from congress.api import webservice
from congress.tests import base


class TestAPIUtils(base.TestCase):

    def setUp(self):
        super(TestAPIUtils, self).setUp()

    def test_create_table_dict(self):
        table_name = 'fake_table'
        schema = {'fake_table': ({'name': 'id', 'desc': None},
                                 {'name': 'name', 'desc': None})}
        expected = {'table_id': table_name,
                    'columns': [{'name': 'id', 'description': None},
                                {'name': 'name', 'description': None}]}
        result = api_utils.create_table_dict(table_name, schema)
        self.assertEqual(expected, result)

    def test_get_id_from_context_ds_id(self):
        context = {'ds_id': 'datasource id'}
        expected = ('datasource-mgr', 'datasource id')
        result = api_utils.get_id_from_context(context,
                                               'datasource-mgr',
                                               'policy-engine')
        self.assertEqual(expected, result)

    def test_get_id_from_context_policy_id(self):
        context = {'policy_id': 'policy id'}
        expected = ('policy-engine', 'policy id')
        result = api_utils.get_id_from_context(context,
                                               'datasource-mgr',
                                               'policy-engine')
        self.assertEqual(expected, result)

    def test_get_id_from_context_with_invalid_context(self):
        context = {'invalid_id': 'invalid id'}

        self.assertRaises(webservice.DataModelException,
                          api_utils.get_id_from_context,
                          context, 'datasource-mgr', 'policy-engine')



# Copyright 2015 Huawei.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import copy
import json
import os

from six.moves import http_client
import webob
import webob.dec

from congress.api import webservice


VERSIONS = {
    "v1": {
        "id": "v1",
        "status": "CURRENT",
        "updated": "2013-08-12T17:42:13Z",
        "links": [
            {
                "rel": "describedby",
                "type": "text/html",
                "href": "http://congress.readthedocs.org/",
            },
        ],
    },
}


def _get_view_builder(request):
    base_url = request.application_url
    return ViewBuilder(base_url)


class ViewBuilder(object):

    def __init__(self, base_url):
        """:param base_url: url of the root wsgi application."""
        self.base_url = base_url

    def build_choices(self, versions, request):
        version_objs = []
        for version in sorted(versions.keys()):
            version = versions[version]
            version_objs.append({
                "id": version['id'],
                "status": version['status'],
                "updated": version['updated'],
                "links": self._build_links(version, request.path),
                })

        return dict(choices=version_objs)

    def build_versions(self, versions):
        version_objs = []
        for version in sorted(versions.keys()):
            version = versions[version]
            version_objs.append({
                "id": version['id'],
                "status": version['status'],
                "updated": version['updated'],
                "links": self._build_links(version),
            })

        return dict(versions=version_objs)

    def build_version(self, version):
        reval = copy.deepcopy(version)
        reval['links'].insert(0, {
            "rel": "self",
            "href": self.base_url.rstrip('/') + '/',
        })
        return dict(version=reval)

    def _build_links(self, version_data, path=None):
        """Generate a container of links that refer to the provided version."""
        href = self._generate_href(version_data['id'], path)

        links = [
            {
                "rel": "self",
                "href": href,
            },
        ]

        return links

    def _generate_href(self, version, path=None):
        """Create an url that refers to a specific version."""
        if path:
            path = path.strip('/')
            return os.path.join(self.base_url, version, path)
        else:
            return os.path.join(self.base_url, version) + '/'


class Versions(object):

    @classmethod
    def factory(cls, global_config, **local_config):
        return cls()

    @webob.dec.wsgify(RequestClass=webob.Request)
    def __call__(self, request):
        """Respond to a request for all Congress API versions."""

        builder = _get_view_builder(request)
        if request.path == '/':
            body = builder.build_versions(VERSIONS)
            status = http_client.OK
        else:
            body = builder.build_choices(VERSIONS, request)
            status = http_client.MULTIPLE_CHOICES

        return webob.Response(body="%s\n" % json.dumps(body),
                              status=status,
                              content_type='application/json')


class VersionV1Handler(webservice.AbstractApiHandler):

    def handle_request(self, request):
        builder = _get_view_builder(request)
        body = builder.build_version(VERSIONS['v1'])
        return webob.Response(body="%s\n" % json.dumps(body),
                              status=http_client.OK,
                              content_type='application/json')



# Copyright (c) 2014 VMware
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_config import cfg
import testtools


class ConfigurationTest(testtools.TestCase):

    def test_defaults(self):
        self.assertEqual('0.0.0.0', cfg.CONF.bind_host)
        self.assertEqual(1789, cfg.CONF.bind_port)
        self.assertFalse(cfg.CONF.tcp_keepalive)
        self.assertEqual(600, cfg.CONF.tcp_keepidle)
        self.assertEqual(1, cfg.CONF.api_workers)
        self.assertEqual('api-paste.ini', cfg.CONF.api_paste_config)
        self.assertEqual('keystone', cfg.CONF.auth_strategy)



# Copyright 2014 VMware.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from congressclient.v1 import client as congress_client
import keystoneclient
from openstack_dashboard.api import base


LITERALS_SEPARATOR = '),'
RULE_SEPARATOR = ':-'
TABLE_SEPARATOR = ':'

LOG = logging.getLogger(__name__)


def _set_id_as_name_if_empty(apidict, length=0):
    try:
        if not apidict._apidict.get('name'):
            id = apidict._apidict['id']
            if length:
                id = id[:length]
                apidict._apidict['name'] = '(%s)' % id
            else:
                apidict._apidict['name'] = id
    except KeyError:
        pass


class PolicyAPIDictWrapper(base.APIDictWrapper):
    def set_id_as_name_if_empty(self):
        _set_id_as_name_if_empty(self)

    def set_id_if_empty(self, id):
        apidict_id = self._apidict.get('id')
        if not apidict_id or apidict_id == "None":
            self._apidict['id'] = id

    def set_value(self, key, value):
        self._apidict[key] = value

    def delete_by_key(self, key):
        del self._apidict[key]


class PolicyRule(PolicyAPIDictWrapper):
    """Wrapper for a Congress policy's rule."""
    def set_id_as_name_if_empty(self):
        # Shorten UUID.
        _set_id_as_name_if_empty(self, length=8)


class PolicyTable(PolicyAPIDictWrapper):
    """Wrapper for a Congress policy's data table."""
    def set_policy_details(self, policy):
        self._apidict['policy_name'] = policy['name']
        self._apidict['policy_owner_id'] = policy['owner_id']


def congressclient(request):
    """Instantiate Congress client."""
    auth_url = base.url_for(request, 'identity')
    user = request.user
    auth = keystoneclient.auth.identity.v2.Token(auth_url, user.token.id,
                                                 tenant_id=user.tenant_id,
                                                 tenant_name=user.tenant_name)
    session = keystoneclient.session.Session(auth=auth)
    region_name = user.services_region

    kwargs = {
        'session': session,
        'auth': None,
        'interface': 'publicURL',
        'service_type': 'policy',
        'region_name': region_name
    }
    return congress_client.Client(**kwargs)


def policies_list(request):
    """List all policies."""
    client = congressclient(request)
    policies_list = client.list_policy()
    results = policies_list['results']
    policies = []
    for p in results:
        policy = PolicyAPIDictWrapper(p)
        # Policies currently have a name but not necessarily a non-"None" id.
        # Use the name to identify the policy, needed to differentiate them in
        # DataTables.
        policy.set_id_if_empty(policy.get('name'))
        policies.append(policy)
    return policies


def policy_create(request, args):
    """Create a policy with the given properties."""
    client = congressclient(request)
    policy = client.create_policy(args)
    return policy


def policy_delete(request, policy_id):
    """Delete a policy by id."""
    client = congressclient(request)
    policy = client.delete_policy(policy_id)
    return policy


def policy_get(request, policy_name):
    """Get a policy by name."""
    # TODO(jwy): Use congress.show_policy() once system policies have unique
    # IDs.
    policies = policies_list(request)
    for p in policies:
        if p['name'] == policy_name:
            return p


def policy_rule_create(request, policy_name, body=None):
    """Create a rule in the given policy, with the given properties."""
    client = congressclient(request)
    rule = client.create_policy_rule(policy_name, body=body)
    return rule


def policy_rule_delete(request, policy_name, rule_id):
    """Delete a rule by id, from the given policy."""
    client = congressclient(request)
    rule = client.delete_policy_rule(policy_name, rule_id)
    return rule


def policy_rules_list(request, policy_name):
    """List all rules in a policy, given by name."""
    client = congressclient(request)
    policy_rules_list = client.list_policy_rules(policy_name)
    results = policy_rules_list['results']
    return [PolicyRule(r) for r in results]


def policy_tables_list(request, policy_name):
    """List all data tables in a policy, given by name."""
    client = congressclient(request)
    policy_tables_list = client.list_policy_tables(policy_name)
    results = policy_tables_list['results']
    return [PolicyTable(t) for t in results]


def policy_table_get(request, policy_name, table_name):
    """Get a policy table in a policy, given by name."""
    client = congressclient(request)
    return client.show_policy_table(policy_name, table_name)


def policy_rows_list(request, policy_name, table_name):
    """List all rows in a policy's data table, given by name."""
    client = congressclient(request)
    policy_rows_list = client.list_policy_rows(policy_name, table_name)
    results = policy_rows_list['results']

    policy_rows = []
    # Policy table rows currently don't have ids. However, the DataTable object
    # requires an id for the table to get rendered properly. Otherwise, the
    # same contents are displayed for every row in the table. Assign the rows
    # ids here.
    id = 0
    for row in results:
        new_row = PolicyAPIDictWrapper(row)
        new_row.set_id_if_empty(id)
        id += 1
        policy_rows.append(new_row)
    return policy_rows


def policy_table_schema_get(request, policy_name, table_name):
    """Get the schema for a policy table, based on the first matching rule."""
    column_names = []
    rules = policy_rules_list(request, policy_name)
    # There might be multiple rules that use the same name in the head. Pick
    # the first matching one, which is what the policy engine currently does.
    for rule in rules:
        rule_def = rule['rule']
        head, _ = rule_def.split(RULE_SEPARATOR)
        if head.strip().startswith('%s(' % table_name):
            start = head.index('(') + 1
            end = head.index(')')
            column_names = head[start:end].split(',')
            break

    schema = {'table_id': table_name}
    schema['columns'] = [{'name': name.strip(), 'description': None}
                         for name in column_names]
    return schema


def datasources_list(request):
    """List all the data sources."""
    client = congressclient(request)
    datasources_list = client.list_datasources()
    datasources = datasources_list['results']
    return [PolicyAPIDictWrapper(d) for d in datasources]


def datasource_get(request, datasource_id):
    """Get a data source by id."""
    # TODO(jwy): Need API in congress_client to retrieve data source by id.
    datasources = datasources_list(request)
    for d in datasources:
        if d['id'] == datasource_id:
            return d


def datasource_get_by_name(request, datasource_name):
    """Get a data source by name."""
    datasources = datasources_list(request)
    for d in datasources:
        if d['name'] == datasource_name:
            return d


def datasource_tables_list(request, datasource_id):
    """List all data tables in a data source, given by id."""
    client = congressclient(request)
    datasource_tables_list = client.list_datasource_tables(datasource_id)
    results = datasource_tables_list['results']
    return [PolicyAPIDictWrapper(t) for t in results]


def datasource_rows_list(request, datasource_id, table_name):
    """List all rows in a data source's data table, given by id."""
    client = congressclient(request)
    datasource_rows_list = client.list_datasource_rows(datasource_id,
                                                       table_name)
    results = datasource_rows_list['results']
    datasource_rows = []
    id = 0
    for row in results:
        new_row = PolicyAPIDictWrapper(row)
        new_row.set_id_if_empty(id)
        id += 1
        datasource_rows.append(new_row)
    return datasource_rows


def datasource_schema_get(request, datasource_id):
    """Get the schema for all tables in the given data source."""
    client = congressclient(request)
    return client.show_datasource_schema(datasource_id)


def datasource_table_schema_get(request, datasource_id, table_name):
    """Get the schema for a data source table."""
    client = congressclient(request)
    return client.show_datasource_table_schema(datasource_id, table_name)


def datasource_table_schema_get_by_name(request, datasource_name, table_name):
    """Get the schema for a data source table."""
    datasource = datasource_get_by_name(request, datasource_name)
    client = congressclient(request)
    return client.show_datasource_table_schema(datasource['id'], table_name)


def datasource_statuses_list(request):
    client = congressclient(request)
    datasources_list = client.list_datasources()
    datasources = datasources_list['results']
    ds_status = []

    for ds in datasources:
        try:
            status = client.list_datasource_status(ds['id'])
        except Exception as e:
            LOG.info("Exception while getting the status: %s" % e)
            status = "not available"
            raise e
        wrapper = PolicyAPIDictWrapper(ds)
        wrapper.set_value('service', ds['name'])
        for key in status:
            value = status[key]
            wrapper.set_value(key, value)
        ds_status.append(wrapper)
    return ds_status



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import time

from oslo_log import log as logging
from tempest import clients
from tempest import config
from tempest.lib import exceptions
from tempest import test

from congress_tempest_tests.tests.scenario import helper
from congress_tempest_tests.tests.scenario import manager_congress

CONF = config.CONF
LOG = logging.getLogger(__name__)


class TestNeutronV2Driver(manager_congress.ScenarioPolicyBase):

    @classmethod
    def skip_checks(cls):
        super(TestNeutronV2Driver, cls).skip_checks()
        if not (CONF.network.tenant_networks_reachable
                or CONF.network.public_network_id):
            msg = ('Either tenant_networks_reachable must be "true", or '
                   'public_network_id must be defined.')
            cls.enabled = False
            raise cls.skipException(msg)

        if not CONF.service_available.neutron:
            skip_msg = ("%s skipped as neutron is not available"
                        % cls.__name__)
            raise cls.skipException(skip_msg)

    def setUp(cls):
        super(TestNeutronV2Driver, cls).setUp()
        cls.os = clients.Manager(cls.admin_manager.auth_provider.credentials)
        cls.networks_client = cls.os.networks_client
        cls.subnets_client = cls.os.subnets_client
        cls.ports_client = cls.os.ports_client
        cls.security_groups_client = cls.os.security_groups_client
        cls.routers_client = cls.os.routers_client
        cls.datasource_id = manager_congress.get_datasource_id(
            cls.admin_manager.congress_client, 'neutronv2')

    @test.attr(type='smoke')
    @test.services('network')
    def test_neutronv2_networks_table(self):

        @helper.retry_on_exception
        def _check_data():
            networks = self.networks_client.list_networks()
            network_map = {}
            for network in networks['networks']:
                network_map[network['id']] = network

            client = self.admin_manager.congress_client
            client.request_refresh(self.datasource_id)
            time.sleep(1)

            network_schema = (client.show_datasource_table_schema(
                self.datasource_id, 'networks')['columns'])

            results = (client.list_datasource_rows(
                self.datasource_id, 'networks'))
            for row in results['results']:
                network_row = network_map[row['data'][0]]
                for index in range(len(network_schema)):
                    if (str(row['data'][index]) !=
                            str(network_row[network_schema[index]['name']])):
                        return False
            return True

        if not test.call_until_true(func=_check_data,
                                    duration=200, sleep_for=10):
            raise exceptions.TimeoutException("Data did not converge in time "
                                              "or failure in server")

    @test.attr(type='smoke')
    @test.services('network')
    def test_neutronv2_ports_tables(self):
        port_schema = (
            self.admin_manager.congress_client.show_datasource_table_schema(
                self.datasource_id, 'ports')['columns'])

        port_sec_binding_schema = (
            self.admin_manager.congress_client.show_datasource_table_schema(
                self.datasource_id, 'security_group_port_bindings')['columns'])

        fixed_ips_schema = (
            self.admin_manager.congress_client.show_datasource_table_schema(
                self.datasource_id, 'fixed_ips')['columns'])

        @helper.retry_on_exception
        def _check_data():
            ports_from_neutron = self.ports_client.list_ports()
            port_map = {}
            for port in ports_from_neutron['ports']:
                port_map[port['id']] = port

            client = self.admin_manager.congress_client
            client.request_refresh(self.datasource_id)
            time.sleep(1)

            ports = (client.list_datasource_rows(self.datasource_id, 'ports'))
            security_group_port_bindings = (
                client.list_datasource_rows(
                    self.datasource_id, 'security_group_port_bindings'))
            fixed_ips = (
                client.list_datasource_rows(self.datasource_id, 'fixed_ips'))

            # Validate ports table
            for row in ports['results']:
                port_row = port_map[row['data'][0]]
                for index in range(len(port_schema)):
                    if (str(row['data'][index]) !=
                            str(port_row[port_schema[index]['name']])):
                        return False

            # validate security_group_port_bindings table
            for row in security_group_port_bindings['results']:
                port_row = port_map[row['data'][0]]
                for index in range(len(port_sec_binding_schema)):
                    row_index = port_sec_binding_schema[index]['name']
                    # Translate port_id -> id
                    if row_index == 'port_id':
                        if (str(row['data'][index]) !=
                                str(port_row['id'])):
                            return False
                    elif row_index == 'security_group_id':
                        if (str(row['data'][index]) not in
                                port_row['security_groups']):
                            return False

            # validate fixed_ips
            for row in fixed_ips['results']:
                port_row = port_map[row['data'][0]]
                for index in range(len(fixed_ips_schema)):
                    row_index = fixed_ips_schema[index]['name']
                    if row_index in ['subnet_id', 'ip_address']:
                        if not port_row['fixed_ips']:
                            continue
                        for fixed_ip in port_row['fixed_ips']:
                            if row['data'][index] == fixed_ip[row_index]:
                                break
                        else:
                            # no subnet_id/ip_address match found
                            return False
            return True

        if not test.call_until_true(func=_check_data,
                                    duration=200, sleep_for=10):
            raise exceptions.TimeoutException("Data did not converge in time "
                                              "or failure in server")

    @test.attr(type='smoke')
    @test.services('network')
    def test_neutronv2_subnets_tables(self):
        subnet_schema = (
            self.admin_manager.congress_client.show_datasource_table_schema(
                self.datasource_id, 'subnets')['columns'])

        host_routes_schema = (
            self.admin_manager.congress_client.show_datasource_table_schema(
                self.datasource_id, 'host_routes')['columns'])

        dns_nameservers_schema = (
            self.admin_manager.congress_client.show_datasource_table_schema(
                self.datasource_id, 'dns_nameservers')['columns'])

        allocation_pools_schema = (
            self.admin_manager.congress_client.show_datasource_table_schema(
                self.datasource_id, 'allocation_pools')['columns'])

        @helper.retry_on_exception
        def _check_data():
            subnets_from_neutron = self.subnets_client.list_subnets()
            subnet_map = {}
            for subnet in subnets_from_neutron['subnets']:
                subnet_map[subnet['id']] = subnet

            client = self.admin_manager.congress_client
            client.request_refresh(self.datasource_id)
            time.sleep(1)

            subnets = (
                client.list_datasource_rows(self.datasource_id, 'subnets'))
            host_routes = (
                client.list_datasource_rows(self.datasource_id, 'host_routes'))
            dns_nameservers = (
                client.list_datasource_rows(
                    self.datasource_id, 'dns_nameservers'))
            allocation_pools = (
                client.list_datasource_rows(
                    self.datasource_id, 'allocation_pools'))
            # Validate subnets table
            for row in subnets['results']:
                subnet_row = subnet_map[row['data'][0]]
                for index in range(len(subnet_schema)):
                    if (str(row['data'][index]) !=
                            str(subnet_row[subnet_schema[index]['name']])):
                        return False

            # validate dns_nameservers
            for row in dns_nameservers['results']:
                subnet_row = subnet_map[row['data'][0]]
                for index in range(len(dns_nameservers_schema)):
                    row_index = dns_nameservers_schema[index]['name']
                    if row_index in ['dns_nameserver']:
                        if (row['data'][index]
                                not in subnet_row['dns_nameservers']):
                            return False

            # validate host_routes
            for row in host_routes['results']:
                subnet_row = subnet_map[row['data'][0]]
                for index in range(len(host_routes_schema)):
                    row_index = host_routes_schema[index]['name']
                    if row_index in ['destination', 'nexthop']:
                        if not subnet_row['host_routes']:
                            continue
                        for host_route in subnet_row['host_routes']:
                            if row['data'][index] == host_route[row_index]:
                                break
                        else:
                            # no destination/nexthop match found
                            return False

            # validate allocation_pools
            for row in allocation_pools['results']:
                subnet_row = subnet_map[row['data'][0]]
                for index in range(len(allocation_pools_schema)):
                    row_index = allocation_pools_schema[index]['name']
                    if row_index in ['start', 'end']:
                        if not subnet_row['allocation_pools']:
                            continue
                        for allocation_pool in subnet_row['allocation_pools']:
                            if (row['data'][index] ==
                                    allocation_pool[row_index]):
                                break
                        else:
                            # no destination/nexthop match found
                            return False
            return True

        if not test.call_until_true(func=_check_data,
                                    duration=200, sleep_for=10):
            raise exceptions.TimeoutException("Data did not converge in time "
                                              "or failure in server")

    @test.attr(type='smoke')
    @test.services('network')
    def test_neutronv2_routers_tables(self):
        router_schema = (
            self.admin_manager.congress_client.show_datasource_table_schema(
                self.datasource_id, 'routers')['columns'])

        ext_gw_info_schema = (
            self.admin_manager.congress_client.show_datasource_table_schema(
                self.datasource_id, 'external_gateway_infos')['columns'])

        @helper.retry_on_exception
        def _check_data():
            routers_from_neutron = self.routers_client.list_routers()
            router_map = {}
            for router in routers_from_neutron['routers']:
                router_map[router['id']] = router

            client = self.admin_manager.congress_client
            client.request_refresh(self.datasource_id)
            time.sleep(1)

            routers = (
                client.list_datasource_rows(self.datasource_id, 'routers'))

            ext_gw_info = (
                client.list_datasource_rows(
                    self.datasource_id, 'external_gateway_infos'))

            # Validate routers table
            for row in routers['results']:
                router_row = router_map[row['data'][0]]
                for index in range(len(router_schema)):
                    if (str(row['data'][index]) !=
                            str(router_row[router_schema[index]['name']])):
                        return False

            # validate external_gateway_infos
            for row in ext_gw_info['results']:
                router_ext_gw_info = (
                    router_map[row['data'][0]]['external_gateway_info'])
                # populate router_id
                router_ext_gw_info['router_id'] = row['data'][0]
                for index in range(len(ext_gw_info_schema)):
                    val = router_ext_gw_info[ext_gw_info_schema[index]['name']]
                    if (str(row['data'][index]) != str(val)):
                        return False
            return True

        if not test.call_until_true(func=_check_data,
                                    duration=200, sleep_for=10):
            raise exceptions.TimeoutException("Data did not converge in time "
                                              "or failure in server")

    @test.attr(type='smoke')
    @test.services('network')
    def test_neutronv2_security_groups_table(self):
        sg_schema = (
            self.admin_manager.congress_client.show_datasource_table_schema(
                self.datasource_id, 'security_groups')['columns'])

        @helper.retry_on_exception
        def _check_data():
            client = self.security_groups_client
            security_groups_neutron = client.list_security_groups()
            security_groups_map = {}
            for security_group in security_groups_neutron['security_groups']:
                security_groups_map[security_group['id']] = security_group

            client = self.admin_manager.congress_client
            client.request_refresh(self.datasource_id)
            time.sleep(1)

            security_groups = (
                client.list_datasource_rows(
                    self.datasource_id, 'security_groups'))

            # Validate security_group table
            for row in security_groups['results']:
                sg_row = security_groups_map[row['data'][0]]
                for index in range(len(sg_schema)):
                    if (str(row['data'][index]) !=
                            str(sg_row[sg_schema[index]['name']])):
                        return False
            return True

        if not test.call_until_true(func=_check_data,
                                    duration=200, sleep_for=10):
            raise exceptions.TimeoutException("Data did not converge in time "
                                              "or failure in server")

    @test.attr(type='smoke')
    @test.services('network')
    def test_neutronv2_security_group_rules_table(self):
        sgrs_schema = (
            self.admin_manager.congress_client.show_datasource_table_schema(
                self.datasource_id, 'security_group_rules')['columns'])

        @helper.retry_on_exception
        def _check_data():
            client = self.security_groups_client
            security_groups_neutron = client.list_security_groups()
            sgrs_map = {}  # security_group_rules
            for sg in security_groups_neutron['security_groups']:
                for sgr in sg['security_group_rules']:
                    sgrs_map[sgr['id']] = sgr

            client = self.admin_manager.congress_client
            client.request_refresh(self.datasource_id)
            time.sleep(1)

            security_group_rules = (
                client.list_datasource_rows(
                    self.datasource_id, 'security_group_rules'))

            # Validate security_group_rules table
            for row in security_group_rules['results']:
                sg_rule_row = sgrs_map[row['data'][1]]
                for index in range(len(sgrs_schema)):
                    if (str(row['data'][index]) !=
                            str(sg_rule_row[sgrs_schema[index]['name']])):
                        return False
            return True

        if not test.call_until_true(func=_check_data,
                                    duration=200, sleep_for=10):
            raise exceptions.TimeoutException("Data did not converge in time "
                                              "or failure in server")



# Copyright (c) 2014 Marist SDN Innovation lab Joint with Plexxi Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import


class MockAffinity(object):
    def __init__(self, uuid, name):
        self.uuid = uuid
        self.name = name

    def getUuid(self):
        return self.uuid

    def getName(self):
        return self.name


class MockCoreSession(object):
    def __init__(self):
        pass

    def disconnect():
        pass


class MockHost(object):
    def __init__(self, uuid, name, mac_count, pnics):
        self.uuid = uuid
        self.name = name
        self.mac_count = mac_count
        self.pnics = pnics
        self.vms = []

    def addvm(self, vm):
        self.vms.append(vm)

    def getForeignUuid(self):
        return self.uuid

    def getUuid(self):
        return self.uuid

    def getName(self):
        return self.name

    def getPhysicalNetworkInterfaces(self):
        return self.pnics

    def getVirtualMachineCount(self):
        return len(self.vms)

    def getVirtualMachines(self):
        return self.vms


class MockNetworkLink(object):
    def __init__(self, uuid, name, stopint, startint):
        self.uuid = uuid
        self.name = name
        self.startint = startint
        self.stopint = stopint

    def getUuid(self):
        return self.uuid

    def getName(self):
        return self.name

    def getStartNetworkInterface(self):
        return self.startint

    def getStopNetworkInterface(self):
        return self.stopint


class MockNIC(object):
    def __init__(self, uuid, mac):
        self.uuid = uuid
        self.mac = mac

    def getMacAddress(self):
        return self.mac


class MockPort(object):
    def __init__(self, uuid, name, networklinks):
        self.uuid = uuid
        self.name = name
        self.networklinks = networklinks

    def getUuid(self):
        return self.uuid

    def getName(self):
        return self.name

    def getNetworkLinks(self):
        return self.networklinks


class MockSwitch(object):
    def __init__(self, uuid, ip, name, status, pnics):
        self.uuid = uuid
        self.ip = ip
        self.status = status
        self.pnics = pnics
        self.name = name

    def getUuid(self):
        return self.uuid

    def getName(self):
        return self.name

    def getIpAddress(self):
        return self.ip

    def getStatus(self):
        return self.status

    def getPhysicalNetworkInterfaces(self):
        return self.pnics


class MockVM(object):
    def __init__(self, uuid, ip, name, host, vnics):
        self.uuid = uuid
        self.ip = ip
        self.host = host
        self.name = name
        self.vnics = vnics

    def getForeignUuid(self):
        return self.uuid

    def getVirtualizationHost(self):
        return self.host

    def getName(self):
        return self.name

    def getIpAddress(self):
        return self.ip

    def getVirtualNetworkInterfaces(self):
        return self.vnics


class MockVSwitch(object):
    def __init__(self, uuid, hosts, vnics):
        self.uuid = uuid
        self.hosts = hosts
        self.vnics = vnics

    def getForeignUuid(self):
        return self.uuid

    def getVirtualizationHosts(self):
        return self.hosts

    def getVirtualNetworkInterfaces(self):
        return self.vnics



# Copyright (c) 2016 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import functools
import json
import time

import eventlet
eventlet.monkey_patch()

from oslo_log import log as logging

from congress.dse2.data_service import DataService

LOG = logging.getLogger()


def drop_cast_echos(wrapped):
    @functools.wraps(wrapped)
    def wrapper(rpc_endpoint, message_context, *args, **kwargs):
        node = rpc_endpoint.dse_bus.node
        if message_context['node_id'] == node.node_id:
            LOG.trace("<%s> Ignoring my echo", node.node_id)
            return
        return wrapped(rpc_endpoint, message_context, *args, **kwargs)
    return wrapper


class HeartbeatEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return 0  # suppress sets
        # Let the base class default method handle all other cases
        return json.JSONEncoder.default(self, obj)


class _DseControlBusEndpoint(object):
    def __init__(self, dse_bus):
        self.dse_bus = dse_bus

    @drop_cast_echos
    def accept_heartbeat(self, client_ctxt, args):
        LOG.debug("<%s> Accepted heartbeat: context=%s, args='%s'",
                  self.dse_bus.node.node_id, client_ctxt, args)
        hb = json.loads(args)
        # convert dict to set
        for target in hb['subscribed_tables']:
            hb['subscribed_tables'][target] = set(
                hb['subscribed_tables'][target])
        peer_id = client_ctxt['node_id']
        new_status = {
            'node_id': peer_id,
            'instance': client_ctxt['instance'],
            'services': hb['services'],
            'subscribed_tables': hb['subscribed_tables']
        }

        old_status = self.dse_bus.peers.get(peer_id)
        if old_status:
            # TODO(pballand): validate instance, services
            LOG.trace("<%s> Refreshed peer '%s' with services %s",
                      self.dse_bus.node.node_id, peer_id,
                      [s['service_id'] for s in new_status['services']])
        else:
            LOG.info("<%s> New peer '%s' with services %s",
                     self.dse_bus.node.node_id, peer_id,
                     [s['service_id'] for s in new_status['services']])
        self.dse_bus.peers[peer_id] = new_status

        # TODO(pballand): handle time going backwards
        self.dse_bus.peers[peer_id]['last_hb_time'] = time.time()

    @drop_cast_echos
    def list_services(self, client_ctxt):
        LOG.debug("<%s> Peer '%s' requested updated service list",
                  self.dse_bus.node.node_id,  client_ctxt['node_id'])
        self.dse_bus._publish_heartbeat()


class DseNodeControlBus(DataService):
    """Maintain DSE connection for a DseNode.

    The DSE maintains a common directory of data services and their
    corresponding exported tables and RPCs.  This control bus maintains
    this view using oslo.messaging RPC primitives.
    """
    HEARTBEAT_INTERVAL = 1

    def __init__(self, node):
        self.node = node
        self.control_bus_ep = _DseControlBusEndpoint(self)
        self.peers = {}
        super(DseNodeControlBus, self).__init__('_control_bus')

    def rpc_endpoints(self):
        return [self.control_bus_ep]

    def _publish_heartbeat(self):
        args = json.dumps(
            {'services': [s.info.to_dict()
                          for s in self.node.get_services(True)],
             'subscribed_tables': self.node.subscriptions},
            cls=HeartbeatEncoder)
        self.node.broadcast_service_rpc(self.service_id, 'accept_heartbeat',
                                        args=args)

    def _heartbeat_loop(self):
        while self._running:
            self._publish_heartbeat()
            self.node._update_tables_with_subscriber()
            eventlet.sleep(self.HEARTBEAT_INTERVAL)

    def _refresh_peers(self):
        # Request immediate status refresh from peers
        LOG.debug("<%s> Requesting service list from all peers",
                  self.node.node_id)
        self.node.broadcast_service_rpc(self.service_id, 'list_services')

    def start(self):
        LOG.debug("<%s> Starting DSE control bus", self.node.node_id)
        super(DseNodeControlBus, self).start()

        # TODO(pballand): ensure I am not currently running
        #  Add an instance UUID to the node status, have timeout on nodes
        self._refresh_peers()

        # TODO(pballand): before enabling self, check if my node ID is
        # already present (no consensus service, so use timeout heuristic)
        eventlet.spawn(self._heartbeat_loop)

    def dse_status(self):
        """Return latest observation of DSE status."""
        # TODO(pballand): include node status [JOINING, JOINED]
        return {'peers': self.peers}



# Copyright (c) 2014 Marist SDN Innovation lab Joint with Plexxi Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import mock

from congress.datasources import vCenter_driver
from congress.tests import base
from congress.tests.datasources import vCenter_fakes
from congress.tests import helper


class TestvCenterDriver(base.TestCase):
    def setUp(self):
        super(TestvCenterDriver, self).setUp()
        args = helper.datasource_openstack_args()
        args['max_hosts'] = 999
        args['max_vms'] = 999
        self.driver = vCenter_driver.VCenterDriver(args=args,
                                                   session="Testing")
        self.mock_rawhosts = {}
        h1_obj = {}
        h1_obj['value'] = 'Host1'
        mock_host1 = vCenter_fakes.MockvCenterHost(h1_obj)
        h1_uuid = vCenter_fakes.MockProperty(
            '9912c61d-79e0-4423-bb43-d79926e0d1f0',
            'hardware.systemInfo.uuid')
        h1_name = vCenter_fakes.MockProperty('Host1', 'name')
        h1_DNS_obj = vCenter_fakes.MockDNSInfo(['10.11.12.1', '10.11.12.2'])
        h1_DNS = vCenter_fakes.MockProperty(h1_DNS_obj,
                                            'config.network.dnsConfig.address')
        h1_pnic1 = {}
        h1_pnic1['uuid'] = '9912c61d-79e0-4423-bb43-d79926e0d1f1'
        h1_pnic1['mac'] = '3F-0B-DD-8A-F3-B9'
        h1_pnic1['device'] = 'vmnic1'
        h1_pnic1['spec'] = {}
        h1_pnic1['spec']['ip'] = {}
        h1_pnic1['spec']['ip']['ipAddress'] = '10.11.13.1'
        h1_pnic1['spec']['ip']['subnetMask'] = '255.255.255.0'
        h1_pnic2 = {}
        h1_pnic2['uuid'] = '9912c61d-79e0-4423-bb43-d79926e0d1f2'
        h1_pnic2['mac'] = '3F-0B-DD-8A-F3-BA'
        h1_pnic2['device'] = 'vmnic2'
        h1_pnic2['spec'] = {}
        h1_pnic2['spec']['ip'] = {}
        h1_pnic2['spec']['ip']['ipAddress'] = '10.11.13.2'
        h1_pnic2['spec']['ip']['subnetMask'] = '255.255.255.0'
        h1_pnic_list = (h1_pnic1, h1_pnic2)
        h1_pnic_obj = vCenter_fakes.MockNicContainer(h1_pnic_list)
        h1_pnics = vCenter_fakes.MockProperty(h1_pnic_obj,
                                              'config.network.pnic')
        h1_vnic1 = {}
        h1_vnic1['uuid'] = '9912c61d-79e0-4423-bb43-d79926e0d1f3'
        h1_vnic1['device'] = 'vmk1'
        h1_vnic1['portgroup'] = 'Management'
        h1_vnic1['spec'] = {}
        h1_vnic1['spec']['mac'] = '3F-0B-DD-8A-F3-BB'
        h1_vnic1['spec']['ip'] = {}
        h1_vnic1['spec']['ip']['ipAddress'] = '10.11.13.3'
        h1_vnic1['spec']['ip']['subnetMask'] = '255.255.255.0'
        h1_vnic2 = {}
        h1_vnic2['uuid'] = '9912c61d-79e0-4423-bb43-d79926e0d1f4'
        h1_vnic2['device'] = 'vmk2'
        h1_vnic2['portgroup'] = 'Public'
        h1_vnic2['spec'] = {}
        h1_vnic2['spec']['mac'] = '3F-0B-DD-8A-F3-BC'
        h1_vnic2['spec']['ip'] = {}
        h1_vnic2['spec']['ip']['ipAddress'] = '10.11.13.4'
        h1_vnic2['spec']['ip']['subnetMask'] = '255.255.255.0'
        h1_vnic_list = [h1_vnic1, h1_vnic2]
        h1_vnic_obj = vCenter_fakes.MockNicContainer(h1_vnic_list)
        h1_vnics = vCenter_fakes.MockProperty(h1_vnic_obj,
                                              'config.network.vnic')
        mock_host1['propSet'] = [h1_uuid, h1_name, h1_DNS, h1_pnics, h1_vnics]
        h2_obj = {}
        h2_obj['value'] = 'Host2'
        mock_host2 = vCenter_fakes.MockvCenterHost(h2_obj)
        h2_uuid = vCenter_fakes.MockProperty(
            '9912c61d-79e0-4423-bb43-d79926e0d1f5',
            'hardware.systemInfo.uuid')
        h2_name = vCenter_fakes.MockProperty('Host2', 'name')
        h2_DNS_obj = vCenter_fakes.MockDNSInfo(['10.11.12.1', '10.11.12.2'])
        h2_DNS = vCenter_fakes.MockProperty(h2_DNS_obj,
                                            'config.network.dnsConfig.address')
        h2_pnic1 = {}
        h2_pnic1['uuid'] = '9912c61d-79e0-4423-bb43-d79926e0d1f6'
        h2_pnic1['mac'] = '3F-0B-DD-8A-F3-BD'
        h2_pnic1['device'] = 'vmnic1'
        h2_pnic1['spec'] = {}
        h2_pnic1['spec']['ip'] = {}
        h2_pnic1['spec']['ip']['ipAddress'] = '10.11.14.1'
        h2_pnic1['spec']['ip']['subnetMask'] = '255.255.255.0'
        h2_pnic2 = {}
        h2_pnic2['uuid'] = '9912c61d-79e0-4423-bb43-d79926e0d1f7'
        h2_pnic2['mac'] = '3F-0B-DD-8A-F3-BE'
        h2_pnic2['device'] = 'vmnic2'
        h2_pnic2['spec'] = {}
        h2_pnic2['spec']['ip'] = {}
        h2_pnic2['spec']['ip']['ipAddress'] = '10.11.14.2'
        h2_pnic2['spec']['ip']['subnetMask'] = '255.255.255.0'
        h2_pnic_list = (h2_pnic1, h2_pnic2)
        h2_pnic_obj = vCenter_fakes.MockNicContainer(h2_pnic_list)
        h2_pnics = vCenter_fakes.MockProperty(h2_pnic_obj,
                                              'config.network.pnic')
        h2_vnic1 = {}
        h2_vnic1['uuid'] = '9912c61d-79e0-4423-bb43-d79926e0d1f8'
        h2_vnic1['device'] = 'vmk1'
        h2_vnic1['portgroup'] = 'Management'
        h2_vnic1['spec'] = {}
        h2_vnic1['spec']['mac'] = '3F-0B-DD-8A-F3-BF'
        h2_vnic1['spec']['ip'] = {}
        h2_vnic1['spec']['ip']['ipAddress'] = '10.11.14.3'
        h2_vnic1['spec']['ip']['subnetMask'] = '255.255.255.0'
        h2_vnic2 = {}
        h2_vnic2['uuid'] = '9912c61d-79e0-4423-bb43-d79926e0d1f9'
        h2_vnic2['device'] = 'vmk2'
        h2_vnic2['portgroup'] = 'Public'
        h2_vnic2['spec'] = {}
        h2_vnic2['spec']['mac'] = '3F-0B-DD-8A-F3-C0'
        h2_vnic2['spec']['ip'] = {}
        h2_vnic2['spec']['ip']['ipAddress'] = '10.11.14.4'
        h2_vnic2['spec']['ip']['subnetMask'] = '255.255.255.0'
        h2_vnic_list = [h2_vnic1, h2_vnic2]
        h2_vnic_obj = vCenter_fakes.MockNicContainer(h2_vnic_list)
        h2_vnics = vCenter_fakes.MockProperty(h2_vnic_obj,
                                              'config.network.vnic')
        mock_host2['propSet'] = [h2_uuid, h2_name, h2_DNS, h2_pnics, h2_vnics]
        mock_hostlist = [mock_host1, mock_host2]
        self.mock_rawhosts['objects'] = mock_hostlist
        self.mock_rawvms = {}
        mock_vm1 = {}
        mock_vm1['value'] = 'VM1'
        vm1_name = vCenter_fakes.MockProperty('VM1', 'name')
        vm1_uuid = vCenter_fakes.MockProperty(
            '9912c61d-79e0-4423-bb43-d79926e0d200',
            'config.uuid')
        vm1_annotation = vCenter_fakes.MockProperty('First VM',
                                                    'config.annotation')
        vm1_path = vCenter_fakes.MockProperty('[Datastore] VM1/VM1.vmtx',
                                              'summary.config.vmPathName')
        vm1_memSize = vCenter_fakes.MockProperty(4096,
                                                 'summary.config.memorySizeMB')
        vm1_status = vCenter_fakes.MockProperty('green',
                                                'summary.overallStatus')
        host1_referance = {}
        host1_referance['value'] = 'Host1'
        vm1_host = vCenter_fakes.MockProperty(host1_referance, 'runtime.host')
        vm1_quickstats = {}
        vm1_quickstats['guestMemoryUsage'] = 245
        vm1_quickstats['overallCpuDemand'] = 216
        vm1_quickstats['overallCpuUsage'] = 192
        vm1_quickstats_property = vCenter_fakes.MockProperty(
            vm1_quickstats, 'summary.quickStats')
        vm1_storage = {}
        vm1_storage['committed'] = 25294964803
        vm1_storage['uncommitted'] = 32812040762
        vm1_storage_property = vCenter_fakes.MockProperty(vm1_storage,
                                                          'summary.storage')
        mock_vm1['propSet'] = [vm1_name, vm1_uuid, vm1_annotation, vm1_path,
                               vm1_memSize, vm1_status, vm1_host,
                               vm1_quickstats_property, vm1_storage_property]
        mock_vm2 = {}
        mock_vm2['value'] = 'VM2'
        vm2_name = vCenter_fakes.MockProperty('VM2', 'name')
        vm2_uuid = vCenter_fakes.MockProperty(
            '9912c61d-79e0-4423-bb43-d79926e0d201',
            'config.uuid')
        vm2_annotation = vCenter_fakes.MockProperty('Second VM',
                                                    'config.annotation')
        vm2_path = vCenter_fakes.MockProperty('[Datastore] VM2/VM2.vmtx',
                                              'summary.config.vmPathName')
        vm2_memSize = vCenter_fakes.MockProperty(4096,
                                                 'summary.config.memorySizeMB')
        vm2_status = vCenter_fakes.MockProperty('green',
                                                'summary.overallStatus')
        host2_referance = {}
        host2_referance['value'] = 'Host2'
        vm2_host = vCenter_fakes.MockProperty(host2_referance, 'runtime.host')
        vm2_quickstats = {}
        vm2_quickstats['guestMemoryUsage'] = 0
        vm2_quickstats['overallCpuDemand'] = 0
        vm2_quickstats['overallCpuUsage'] = 0
        vm2_quickstats_property = vCenter_fakes.MockProperty(
            vm1_quickstats,
            'summary.quickStats')
        vm2_storage = {}
        vm2_storage['committed'] = 6271694636
        vm2_storage['uncommitted'] = 34110177822
        vm2_storage_property = vCenter_fakes.MockProperty(vm2_storage,
                                                          'summary.storage')
        mock_vm2['propSet'] = [vm2_name, vm2_uuid, vm2_annotation, vm2_path,
                               vm2_memSize, vm2_status, vm2_host,
                               vm2_quickstats_property, vm2_storage_property]
        mock_vmlist = [mock_vm1, mock_vm2]
        self.mock_rawvms['objects'] = mock_vmlist

    def test_translators(self):
        with mock.patch.object(self.driver, '_get_hosts_from_vcenter',
                               return_value=self.mock_rawhosts):
            hosts, pnics, vnics = self.driver._get_hosts_and_nics()
        self.driver._translate_hosts(hosts)
        self.driver._translate_pnics(pnics)
        self.driver._translate_vnics(vnics)
        expected_hosts = set([('Host1',
                               '9912c61d-79e0-4423-bb43-d79926e0d1f0',
                               '895f69d340dac8cd4c9550e745703c77'),
                              ('Host2',
                               '9912c61d-79e0-4423-bb43-d79926e0d1f5',
                               '895f69d340dac8cd4c9550e745703c77')])
        self.assertEqual(expected_hosts, self.driver.state['hosts'])
        expected_DNS = set([('895f69d340dac8cd4c9550e745703c77',
                             '10.11.12.1'),
                            ('895f69d340dac8cd4c9550e745703c77',
                             '10.11.12.2')])
        self.assertEqual(expected_DNS, self.driver.state['host.DNS_IPs'])
        expected_pnics = set([('9912c61d-79e0-4423-bb43-d79926e0d1f0',
                               'vmnic1',
                               '3F-0B-DD-8A-F3-B9',
                               '10.11.13.1',
                               '255.255.255.0'),
                              ('9912c61d-79e0-4423-bb43-d79926e0d1f0',
                               'vmnic2',
                               '3F-0B-DD-8A-F3-BA',
                               '10.11.13.2',
                               '255.255.255.0'),
                              ('9912c61d-79e0-4423-bb43-d79926e0d1f5',
                               'vmnic1',
                               '3F-0B-DD-8A-F3-BD',
                               '10.11.14.1',
                               '255.255.255.0'),
                              ('9912c61d-79e0-4423-bb43-d79926e0d1f5',
                               'vmnic2',
                               '3F-0B-DD-8A-F3-BE',
                               '10.11.14.2',
                               '255.255.255.0')])
        self.assertEqual(expected_pnics, self.driver.state['host.PNICs'])
        expected_vnics = set([('9912c61d-79e0-4423-bb43-d79926e0d1f0',
                               'vmk1',
                               '3F-0B-DD-8A-F3-BB',
                               'Management',
                               '10.11.13.3',
                               '255.255.255.0'),
                              ('9912c61d-79e0-4423-bb43-d79926e0d1f0',
                               'vmk2',
                               '3F-0B-DD-8A-F3-BC',
                               'Public',
                               '10.11.13.4',
                               '255.255.255.0'),
                              ('9912c61d-79e0-4423-bb43-d79926e0d1f5',
                               'vmk1',
                               '3F-0B-DD-8A-F3-BF',
                               'Management',
                               '10.11.14.3',
                               '255.255.255.0'),
                              ('9912c61d-79e0-4423-bb43-d79926e0d1f5',
                               'vmk2',
                               '3F-0B-DD-8A-F3-C0',
                               'Public',
                               '10.11.14.4',
                               '255.255.255.0')])
        self.assertEqual(expected_vnics, self.driver.state['host.VNICs'])
        with mock.patch.object(self.driver, '_get_vms_from_vcenter',
                               return_value=self.mock_rawvms):
            vms = self.driver._get_vms()
        self.driver._translate_vms(vms)
        expected_vms = set([('VM1',
                             '9912c61d-79e0-4423-bb43-d79926e0d200',
                             '9912c61d-79e0-4423-bb43-d79926e0d1f0',
                             '[Datastore] VM1/VM1.vmtx',
                             'green',
                             216,
                             192,
                             4096,
                             245,
                             25294964803,
                             32812040762,
                             'First VM'),
                            ('VM2',
                             '9912c61d-79e0-4423-bb43-d79926e0d201',
                             '9912c61d-79e0-4423-bb43-d79926e0d1f5',
                             '[Datastore] VM2/VM2.vmtx',
                             'green',
                             216,
                             192,
                             4096,
                             245,
                             6271694636,
                             34110177822,
                             'Second VM')])
        self.assertEqual(expected_vms, self.driver.state['vms'])

    def test_execute(self):
        class vCenterClient(object):
            def __init__(self):
                self.testkey = None

            def connectNetwork(self, arg1):
                self.testkey = 'arg1=%s' % arg1

        vcenter_client = vCenterClient()
        self.driver.session = vcenter_client
        api_args = {
            'positional': ['1']
        }
        expected_ans = 'arg1=1'

        self.driver.execute('connectNetwork', api_args)

        self.assertEqual(expected_ans, vcenter_client.testkey)



# Copyright 2014 Plexxi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Main entrypoint for the DSE
#
# Configuration in d6cage.ini
#
# Prerequisites:
# - Plexxi API libraries (there is an RPM)
# - Python dependencies (see readme elsewhere, or capture RPM)

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import imp
import sys
import traceback

import eventlet
eventlet.monkey_patch()
from oslo_log import log as logging
from oslo_utils import importutils
from oslo_utils import strutils

from congress.dse import amqprouter
from congress.dse import d6message
from congress.dse import deepsix


LOG = logging.getLogger(__name__)


class DataServiceError (Exception):
    pass


# This holds the cage instance singleton
instances = {}


def singleton(class_):
    global instances

    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]
    return getinstance


def delete_cage():
    global instances
    for instance in instances.values():
        del instance
    instances = {}


@singleton
class d6Cage(deepsix.deepSix):
    def __init__(self):
        self.config = {}
        self.config['modules'] = {}
        self.config['services'] = {}

        # Dictionary mapping service name to a dict of arguments.
        # Those arguments are only passed to d6service by createservice if they
        #   are not alreay present in the ARGS argument given to createservice.
        self.default_service_args = {}

        cageKeys = ['python.d6cage']
        cageDesc = 'deepsix python cage'
        name = "d6cage"

        deepsix.deepSix.__init__(self, name, cageKeys)

        self.inbox = eventlet.Queue()
        self.dataPath = eventlet.Queue()

        self.table = amqprouter.routeTable()
        self.table.add("local.router", self.inbox)
        self.table.add(self.name, self.inbox)
        self.table.add("router", self.inbox)

        localname = "local." + self.name
        self.table.add(localname, self.inbox)

        self.modules = {}
        self.services = {}

        self.greenThreads = []

        self.unloadingServices = {}
        self.reloadingServices = set()

        self.services[self.name] = {}
        self.services[self.name]['service'] = self
        self.services[self.name]['name'] = self.name
        self.services[self.name]['description'] = cageDesc
        self.services[self.name]['inbox'] = self.inbox
        self.services[self.name]['keys'] = self.keys
        self.services[self.name]['type'] = None
        self.services[self.name]['id'] = None

        self.subscribe(
            "local.d6cage",
            "routeKeys",
            callback=self.updateRoutes,
            interval=5)

        self.router_greenthread = eventlet.spawn(self.router_loop)

    def __del__(self):
        # This function gets called when the interpreter deletes the object
        # by the automatic garbage cleanup
        for gt in self.greenThreads:
            eventlet.kill(gt)

        eventlet.kill(self.router_greenthread)
        eventlet.kill(self)

    def newConfig(self, msg):
        newConfig = msg.body.data
        if type(newConfig) == dict and newConfig:
            if "modules" in newConfig:
                for module in newConfig["modules"]:
                    if module not in sys.modules:
                        self.loadModule(
                            module,
                            newConfig['modules'][module]['filename'])

            if "services" in newConfig:
                for service in newConfig['services']:
                    if service not in self.services:
                        self.createservice(
                            service,
                            **newConfig['services'][service])

            self.config = newConfig

    def reloadStoppedService(self, service):
        moduleName = self.config['services'][service]['moduleName']

        try:
            reload(sys.modules[moduleName])
        except Exception as errmsg:
            self.log_error(
                "Unable to reload module '%s': %s", moduleName, errmsg)
            return

        self.createservice(service, **self.config['services'][service])

    def waitForServiceToStop(
            self,
            service,
            attemptsLeft=20,
            callback=None,
            cbkwargs={}):

        if attemptsLeft > 0:

            if self.services[service]['object'].isActive():

                self.timerThreads.append(
                    eventlet.spawn_after(10,
                                         self.waitForServiceToStop,
                                         service,
                                         attemptsLeft - 1))

            else:

                del self.services[service]

                if callback:
                    callback(**cbkwargs)

        else:
            self.log_error("Unable to stop service %s", service)

    def loadModule(self, name, filename):
        if name in sys.modules:
            # self.log_error(
            #     "error loading module '%s': module already exists", name)
            return
        try:
            self.log_info("loading module: %s", name)
            imp.load_source(name, filename)
        except Exception:
            raise DataServiceError(
                "error loading module '%s' from '%s': %s" %
                (name, filename, traceback.format_exc()))

    def load_modules_from_config(self):
        for section in self.config['modules'].keys():
            filename = self.config['modules'][section]["filename"]

            self.loadModule(section, filename)

    def deleteservice(self, name):
        self.log_info("deleting service: %s", name)
        obj = self.services[name]['object']
        if hasattr(obj, "cleanup"):
            obj.cleanup()
        eventlet.greenthread.kill(obj)
        self.greenThreads.remove(obj)
        self.table.remove(name, self.services[name]['inbox'])
        self.table.remove("local." + name, self.services[name]['inbox'])
        self.unsubscribe(name, 'routeKeys')
        del self.services[name]
        self.log_info("finished deleting service: %s", name)

    def createservice(
            self,
            name="",
            keys="",
            description="",
            moduleName="",
            args={},
            module_driver=False,
            type_=None,
            id_=None):

        self.log_info("creating service %s with module %s and args %s",
                      name, moduleName, strutils.mask_password(args, "****"))

        # FIXME(arosen) This will be refactored out in the next patchset
        # this is only done because existing imports from d6service
        # instead of the module.
        if module_driver:
            congress_expected_module_path = ""
            for entry in range(len(moduleName.split(".")) - 1):
                congress_expected_module_path += (
                    moduleName.split(".")[entry] + ".")
            congress_expected_module_path = congress_expected_module_path[:-1]
            module = importutils.import_module(congress_expected_module_path)

        if not module_driver and moduleName not in sys.modules:
            self.log_error(
                "error loading service %s: module %s does not exist",
                name,
                moduleName)
            raise DataServiceError(
                "error loading service %s: module %s does not exist" %
                (name, moduleName))

        if not module_driver and name in self.services:
            self.log_error("error loading service '%s': name already in use",
                           name)
            raise DataServiceError(
                "error loading service '%s': name already in use"
                % name)

        inbox = eventlet.Queue()
        if not module_driver:
            module = sys.modules[moduleName]

        # set args to default values, as necessary
        if name in self.default_service_args:
            global_args = self.default_service_args[name]
            for key, value in global_args.items():
                if key not in args:
                    args[key] = value

        try:
            svcObject = module.d6service(name, keys, inbox, self.dataPath,
                                         args)
            self.greenThreads.append(svcObject)
        except Exception:
            self.log_error("Error loading service '%s' of module '%s':: \n%s",
                           name, module, traceback.format_exc())
            raise DataServiceError(
                "Error loading service '%s' of module '%s':: \n%s"
                % (name, module, traceback.format_exc()))

        self.log_info("created service: %s", name)
        self.services[name] = {}
        self.services[name]['name'] = name
        self.services[name]['description'] = description
        self.services[name]['moduleName'] = moduleName
        self.services[name]['keys'] = keys
        self.services[name]['args'] = args
        self.services[name]['object'] = svcObject
        self.services[name]['inbox'] = inbox
        self.services[name]['type'] = type_
        self.services[name]['id'] = id_

        try:
            self.table.add(name, inbox)
            localname = "local." + name
            self.table.add(localname, inbox)

            self.subscribe(
                name,
                'routeKeys',
                callback=self.updateRoutes,
                interval=5)

            self.publish('services', self.services)
        except Exception as errmsg:
            del self.services[name]
            self.log_error("error starting service '%s': %s", name, errmsg)
            raise DataServiceError(
                "error starting service '%s': %s" % (name, errmsg))

    def getservices(self):
        return self.services

    def getservice(self, id_=None, type_=None, name=None):
        # Returns the first service that matches all non-None parameters.
        for name_, service in self.services.items():
            if (id_ and (not service.get('id', None) or id_ != service['id'])):
                continue
            if type_ and type_ != service['type']:
                continue
            if name and name_ != name:
                continue
            return service
        return None

    def service_object(self, name):
        if name in self.services:
            return self.services[name]['object']
        else:
            return None

    def updateRoutes(self, msg):
        keyData = self.getSubData(msg.correlationId, sender=msg.replyTo)
        currentKeys = set(keyData.data)
        # self.log_debug("updateRoutes msgbody: %s", msg.body.data)
        pubKeys = set(msg.body.data['keys'])

        if currentKeys != pubKeys:

            newKeys = pubKeys - currentKeys

            if newKeys:
                self.table.add(
                    list(newKeys), self.services[msg.replyTo]['inbox'])

            oldKeys = currentKeys - pubKeys

            if oldKeys:
                self.table.remove(
                    list(oldKeys), self.services[msg.replyTo]['inbox'])

            return msg.body

    def load_services_from_config(self):

        for section in self.config['services'].keys():

            self.createservice(section, **self.config['services'][section])

    def routemsg(self, msg):
        # LOG.debug(
        #     "Message lookup %s from %s", msg.key, msg.replyTo)

        destinations = self.table.lookup(msg.key)
        # self.log_debug("Destinations %s for key %s for msg %s",
        #     destinations, msg.key, msg)

        if destinations:
            for destination in destinations:
                destination.put_nowait(msg)
                # self.log_debug("Message sent to %s from %s: %s",
                #                 msg.key, msg.replyTo, msg)

    def d6reload(self, msg):

        inargs = msg.body.data

        service = inargs['service']

        newmsg = d6message.d6msg(key=service, replyTo=self.name, type="shut")

        self.send(newmsg)
        cbkwargs = {}

        cbkwargs['service'] = service

        self.waitForServiceToStop(
            service,
            callback=self.reloadStoppedService,
            cbkwargs=cbkwargs)

    def cmdhandler(self, msg):

        command = msg.header['dataindex']

        if command == "reload":
            self.d6reload(msg)

    def router_loop(self):
        while self.running:
            msg = self.dataPath.get()
            self.routemsg(msg)
            self.dataPath.task_done()


if __name__ == '__main__':
    main = d6Cage

    try:
        main.wait()
        main.d6stop()
    except KeyboardInterrupt:
        main.d6stop()
        sys.exit(0)



# Copyright (c) 2015 Intel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_config import cfg

from congress.api import action_model
from congress.api import webservice
from congress import harness
from congress.managers import datasource as datasource_manager
from congress.tests import base
from congress.tests import helper


class TestActionModel(base.SqlTestCase):
    def setUp(self):
        super(TestActionModel, self).setUp()
        # Here we load the fake driver
        cfg.CONF.set_override(
            'drivers',
            ['congress.tests.fake_datasource.FakeDataSource'])

        # NOTE(arosen): this set of tests, tests to deeply. We don't have
        # any tests currently testing cage. Once we do we should mock out
        # cage so we don't have to create one here.

        self.cage = harness.create(helper.root_path())
        self.datasource_mgr = datasource_manager.DataSourceManager
        self.datasource_mgr.validate_configured_drivers()
        req = {'driver': 'fake_datasource',
               'name': 'fake_datasource'}
        req['config'] = {'auth_url': 'foo',
                         'username': 'foo',
                         'password': 'password',
                         'tenant_name': 'foo'}
        self.datasource = self.datasource_mgr.add_datasource(req)
        engine = self.cage.service_object('engine')
        self.action_model = action_model.ActionsModel(
            "action_schema", {}, policy_engine=engine,
            datasource_mgr=self.datasource_mgr)

    def test_get_datasource_actions(self):
        context = {'ds_id': self.datasource['id']}
        actions = self.action_model.get_items({}, context=context)
        expected_ret = {'results': [{'name': 'fake_act',
                        'args': [{'name': 'server_id',
                                  'description': 'server to act'}],
                        'description': 'fake action'}]}
        self.assertEqual(expected_ret, actions)

    def test_get_invalid_datasource_action(self):
        context = {'ds_id': 'invalid_id'}
        self.assertRaises(webservice.DataModelException,
                          self.action_model.get_items, {}, context=context)
